"""
Python interface to the AMPL modeling language

M. Friedlander, University of British Columbia
D. Orban, Ecole Polytechnique de Montreal
$Id$
"""

import numpy
import _amplpy
from pysparse.pysparseMatrix import PysparseMatrix as sp
import sparse_vector_class as sv
import tempfile, os

__docformat__ = 'restructuredtext'

########################################################################


########################################################################


########################################################################

class AmplModel:

    """
    AmplModel creates an instance of an Ampl model. If the nl file is
    already available, simply call AmplModel(stub) where the string
    stub is the name of the model. For instance: AmplModel('elec').
    If only the .mod file is available, set the positional parameter
    neednl to anything but None so Ampl generates the nl file, as in
    AmplModel('elec', neednl=1).
    """

    # Constructor
    def __init__(self, model, **kwargs):

        data = kwargs.get('data', None)
        opts = kwargs.get('opts', None)
        coords = kwargs.get('coords', 0)  # 1 for coord format, 0 for LL format

        # Generate nl file if requested.
        neednl = 'neednl' in kwargs and kwargs['neednl']
        if neednl:
            template = self.GenTemplate(model, data, opts)
            try:
                writestub(template)
            except:
                raise RuntimeError, 'Error generating nl file.'
            
        # Initialize the Ampl module
        _amplpy.ampl_init(model)

        # Get basic info on problem
        (self.n, self.m) = _amplpy.get_dim()  # nvar and ncon
        self.x0   = _amplpy.get_x0()          # initial primal estimate
        self.pi0  = _amplpy.get_pi0()         # initial dual estimate
        self.Lvar = _amplpy.get_Lvar()        # lower bounds on variables
        self.Uvar = _amplpy.get_Uvar()        # upper bounds on variables
        self.Lcon = _amplpy.get_Lcon()        # lower bounds on constraints
        self.Ucon = _amplpy.get_Ucon()        # upper bounds on constraints
        (self.lin, self.nln, self.net) = _amplpy.get_CType() # Constraint types
        self.nlin = len(self.lin)             # number of linear    constraints
        self.nnln = len(self.nln)             #    ...    nonlinear   ...
        self.nnet = len(self.net)             #    ...    network     ...

        # Get sparsity info
        self.nnzj = _amplpy.get_nnzj()        # number of nonzeros in Jacobian
        self.nnzh = _amplpy.get_nnzh()        #                       Hessian

        # Initialize local value for Infinity
        self.Infinity = numpy.Inf
        self.negInfinity = - numpy.Inf
        self.zero = 0.0
        self.one = 1.0

        # Maintain lists of indices for each type of constraints:
        self.rangeC = []    # Range constraints:       cL <= c(x) <= cU
        self.lowerC = []    # Lower constraints:       cL <= c(x)
        self.upperC = []    # Upper constraints:             c(x) <= cU
        self.equalC = []    # Equality constraints:    cL  = c(x)  = cU
        self.freeC  = []    # "Free" constraints:    -inf <= c(x) <= inf

        for i in range(self.m):
            if self.Lcon[i] > self.negInfinity and self.Ucon[i] < self.Infinity:
                if self.Lcon[i] == self.Ucon[i]:
                    self.equalC.append(i)
                else:
                    self.rangeC.append(i)
            elif self.Lcon[i] > self.negInfinity:
                self.lowerC.append(i)
            elif self.Ucon[i] < self.Infinity:
                self.upperC.append(i)
            else:
                # Normally, we should not get here
                self.freeC.append(i)

        self.nlowerC = len(self.lowerC)
        self.nrangeC = len(self.rangeC)
        self.nupperC = len(self.upperC)
        self.nequalC = len(self.equalC)
        self.nfreeC  = len(self.freeC )

        # Proceed similarly with bound constraints
        self.rangeB = []
        self.lowerB = []
        self.upperB = []
        self.fixedB = []
        self.freeB  = []

        for i in range(self.n):
            if self.Lvar[i] > self.negInfinity and self.Uvar[i] < self.Infinity:
                if self.Lvar[i] == self.Uvar[i]:
                    self.fixedB.append(i)
                else:
                    self.rangeB.append(i)
            elif self.Lvar[i] > self.negInfinity:
                self.lowerB.append(i)
            elif self.Uvar[i] < self.Infinity:
                self.upperB.append(i)
            else:
                # This is a free variable
                self.freeB.append(i)

        self.nlowerB = len(self.lowerB)
        self.nrangeB = len(self.rangeB)
        self.nupperB = len(self.upperB)
        self.nfixedB = len(self.fixedB)
        self.nfreeB  = len(self.freeB )
        self.nbounds = self.n - self.nfreeB

        # Define default stopping tolerances
        self.stop_d = 1.0e-6    # Dual feasibility
        self.stop_c = 1.0e-6    # Complementarty
        self.stop_p = 1.0e-6    # Primal feasibility

        # Set matrix format
        self.coords = coords     # Coordinate or LL format
        self.mformat = int(self.coords)

        # Initialize some counters
        self.feval = 0    # evaluations of objective  function
        self.geval = 0    #                           gradient
        self.Heval = 0    #                Lagrangian Hessian
        self.Hprod = 0    #                matrix-vector products with Hessian
        self.ceval = 0    #                constraint functions
        self.Jeval = 0    #                           gradients
        self.Jprod = 0    #                matrix-vector products with Jacobian

    def GenTemplate(self, model, data = None, opts = None):
        """
        Write out an Ampl template file,
        using files model.mod and data.dat (if available).
        The template will be given a temporary name.
        """

        # Create a temporary template file and write in a header.
        tmpname  = tempfile.mktemp()
        template = open(tmpname, 'w')
        template.write("# Template file for %s.\n" % model)
        template.write("# This file was created by AmplPy.\n")

        # Later we can use opts to hold a list of Ampl options, eg,
        #     option presolve 0;
        # that can be written into the template file.
        if opts is not None:
            pass
    
        # Template file body.
        template.write("model %s.mod;\n"     % model)
        if data is not None:
            template.write("data  %s.dat;\n" % data)
        template.write("write g%s;\n"        % model)

        # Finish off the template file.
        template.write("quit;\n")
        template.close()

        # We'll need to know the template file name.
        return tmpname

    def writestub(template):
        os.system("ampl %s" % template)
        return None

    def ResetCounters(self):
        self.feval = 0
        self.geval = 0
        self.Heval = 0
        self.Hprod = 0
        self.ceval = 0
        self.Jeval = 0
        self.Jprod = 0
        return None

    # Destructor
    def close(self):
        _amplpy.ampl_shut()

    def writesol(self, x, z, msg):
        """
        writesol(x, z, msg):
          Write primal-dual solution and message msg to stub.sol
        """
        return _amplpy.ampl_sol(x, z, msg)    

    def Bounds(self, x):
        """
        Evaluate the vector of bound constraints converted to nonnegativity
        bounds. Bounds appear in natural order. The upper side on two-sided
        bounds appears last.
        """
        lowerB = self.lowerB ; nlowerB = self.nlowerB
        upperB = self.upperB ; nupperB = self.nupperB
        rangeB = self.rangeB ; nrangeB = self.nrangeB
        n  = self.n

        b = numpy.empty(n + nrangeB)
        b[:n] = x[:]
        b[n:] = x[rangeB]

        b[lowerB] -= self.Lvar[lowerB]

        b[upperB] -= self.Uvar[upperB] ; b[upperB] *= -1

        b[rangeB] -= self.Lvar[rangeB]
        b[n:]     -= self.Uvar[rangeB] ; b[n:] *= -1

        return b
        
    # The following method allows checking of optimality conditions

    def OptimalityResiduals(self, x, y=None, z=None, **kwargs):
        """
        OptimalityResiduals(x, y, z):
          Evaluate the KKT residuals at (x, y, z).
           x  is the vector of primal variables,
           y  is the vector of Lagrange multipliers for general constraints,
           z  is the vector of Lagrange multipliers for simple bounds.

          If the problem contains range constraints and/or
          two-sided bounds, the multipliers corresponding
          to those must not be combined in pi and z. Instead, they should be
          passed as additional keyword arguments, as in

          OptimalityResiduals(x, y, z, yu = yu, zu = zu)

          where y   are the multipliers corresponding to equality constraints
                    and one-sided inequalities. For inequality constraints,
                    the sign of the multiplier should be as if the constraint
                    were a nonnegativity constraint, i.e., if the constraint
                        cL[i] <= ci(x)
                    is present, the multiplier y[i] should correspond to
                        ci(x) - cL[i] >= 0.
                    If the constraint
                        ci(x) <= cU[i]
                    is present, the multiplier y[i] should correspond to
                        cU[i] - ci(x) >= 0.
                    
                    If constraint i is a range constraint, i.e., of the form
                        cL[i] <= ci(x) <= cU[i]
                    with -infty < cL[i] < cU[i] < +infty,
                    then y[i] should contains the multiplier corresponding to
                        ci(x) - cL[i] >= 0.
                        
                    y is a numpy array.

                z   are the multipliers corresponding to bound constraints
                    on x. The sign of the multiplier should be as if the
                    bound were a nonnegativity constraint, i.e., if the bound
                        xL[i] <= x_i
                    is present, the multiplier z[i] should correspond to
                        x_i - xL[i] >= 0.
                    If the bound
                        x_i <= xU[i]
                    is present, the multiplier z[i] should correspond to
                        xU[i] - x_i >= 0.
                    If constraint i is a two-sided bound constraint, i.e., if it
                    has the form
                        xL[i] <= xi <= xU[i]
                    with -infty < xL[i] < xU[i] < +infty,
                    then z[i] should contain the multiplier corresponding to
                        xi - xL[i] >= 0.
                        
                    z is a dictionary.

          If the problem containts range constraints,

                yu  should be a dictionary containing multipliers corresponding
                    to the constraint
                            cU[i] - ci(x) >= 0
                    where i is the index of a range constraint, i.e., i should
                    be in yu.keys() and its value should be the multiplier.

          If the problem contains two-sided bounds,

                zu  should be a dictionary containing multipliers corresponding
                    to the bounds
                            xU[i] - x[i] >= 0
                    where i is the index of a two-sided bound, i.e., i should be
                    in zu.keys() and its value should be the multipliers.

          It is possible to check the Fritz-John conditions via the keyword FJ.
          If FJ is present, it should be assigned the multiplier value that will
          be applied to the gradient of the objective function.

          Example:
          
              OptimalityResiduals(x, y, z, FJ = 1.0e-5)

          If FJ has value 0.0, the gradient of the objective will not be
          included in the residuals (it will not be computed).

          Other accepted keywords:

          +------+-----------------------+
          | grad | gradient of objective |
          +------+-----------------------+
          | cons | vector of constraints |
          +------+-----------------------+
          | jac  | Jacobian matrix       |
          +------+-----------------------+

          If grad, cons or jac is not supplied, it will be evaluated.

          The function returns a tuple of unscaled residuals, in infinity norm.
           (dual_feas, compl, bnd_compl, primal_feas, bnd_feas)
        """
        # See if we are checking the Fritz-John conditions
        FJ = ('FJ' in kwargs.keys())
        objMult = kwargs.get('FJ', 0.0)

        # If range constraints/bounds are present, record multipliers
        yu = kwargs.get('yu', {})
        zu = kwargs.get('zu', {})

        # Check that y was given if there are general constraints
        if self.m > 0 and y is None:
            raise TypeError, 'Multipliers must be given for general constraints'

        # Check multipliers signs
        if numpy.any(y[lowerC+upperC+rangeC] < 0) or numpy.any(yu<0):
            msg = 'Multipliers associated to general constraints'
            msg += 'must be positive'
            raise ValueError, msg

        # Check that z was given if there are bound constraints
        nBounds = self.nlowerB + self.nupperB + self.nrangeB
        if nBounds > 0 and z is None:
            raise ValueError, 'Multipliers must be given for bound constraints'

        # Check multipliers signs
        if numpy.any(z < 0) or numpy.any(zu < 0):
            msg = 'Multipliers associated to bound constraints'
            msg += 'must be positive'
            raise ValueError, msg

        # Transfer some pointers for readability
        lowerB = self.lowerB ; upperB = self.upperB ; rangeB = self.rangeB
        Lvar   = self.Lvar   ; Uvar   = self.Uvar

        equalC = self.equalC
        lowerC = self.lowerC ; upperC = self.upperC ; rangeC = self.rangeC
        Lcon   = self.Lcon   ; Ucon   = self.Ucon

        # Bounds feasibility
        bnd = numpy.empty(n + nrangeB)
        bnd[:n] = x.copy()
        bnd[n:] = x[upperB]
        bnd[lowerB+rangeB] -= Lvar[lowerB+rangeB]
        bnd[upperB] *= -1
        bnd[upperB] += Uvar[upperB]
        bnd[n:] *= -1
        bnd[n:] += Uvar[rangeB]

        #bndUpper = max([0.0] + [(x[i] - Uvar[i]) for i in upperB + rangeB])
        #bndLower = max([0.0] + [(Lvar[i] - x[i]) for i in lowerB + rangeB])
        #bndFeas  = max(bndFixed, bndUpper, bndLower)

        # Primal feasibility
        if 'cons' in kwargs.keys():
            pFeas = numpy.empty(m + nrangeC)
            pFeas[:m] = kwargs.get('cons').copy()
            pFeas[m:] = pFeas[rangeC]
            pFeas[equalC] -= Lcon[equalC]
            pFeas[lowerC+rangeC] -= Lcon[lowerC+rangeC]
            pFeas[upperC] *= -1
            pFeas[upperC] += Ucon[upperC]
            pFeas[rangeC] *= -1
            pFeas[rangeC] += Ucon[rangeC]
        else:
            pFeas = self.consPos(x)

        #pFeasEq = max([0.0] + [abs(c[i] - Lcon[i]) for i in equalC])
        #pFeasUp = max([0.0] + [max(c[i] - Ucon[i], 0) for i in upperC + rangeC])
        #pFeasLo = max([0.0] + [max(Lcon[i] - c[i], 0) for i in lowerC + rangeC])
        #pFeas   = max(pFeasEq, pFeasUp, pFeasLo)

        # Bounds complementarity
        bndCmp = numpy.empty(n + nrangeB)
        bndCmp[lowerB+rangeB] = numpy.multiply(bndLower, z[lowerB+rangeB])
        bndCmp[upperB] = numpy.multiply(bndUpper, z[upperB])
        bndCmp[n:] = numpy.multiply(bnd[n:], z[n:])

        #bndComp = 0.0
        #for i in lowerB + rangeB:
        #    if z[i] < 0.0:
        #        raise ValueError, 'Negative multipliers: z[%d] = %-g' % (i,z[i])
        #    bndComp = max(bndComp, (x[i] - Lvar[i]) * z[i])
        #for i in upperB:
        #    if z[i] < 0.0:
        #        raise ValueError, 'Negative multipliers: z[%d] = %-g' % (i,z[i])
        #    bndComp = max(bndComp, (Uvar[i] - x[i]) * z[i])
        #for i in rangeB:
        #    if zu[i] < 0.0:
        #        msg = 'Negative multipliers: zu[%d] = %-g' % (i,zu[i])
        #        raise ValueError, msg
        #    bndComp = max(bndComp, (Uvar[i] - x[i]) * zu[i])

        # General constraints complementarity
        comp = 0.0
        for i in lowerC + rangeC:
            if y[i] < 0.0:
                raise ValueError, 'Negative multipliers: y[%d] = %-g' % (i,y[i])
            comp = max(comp, (c[i] - Lcon[i]) * y[i])
        for i in upperC:
            if y[i] < 0.0: 
                raise ValueError, 'Negative multipliers: y[%d] = %-g' % (i,y[i])
                comp = max(comp, (Ucon[i] - c[i]) * y[i])
        for i in rangeC:
            if yu[i] < 0.0:
                raise ValueError, 'Negative multipliers: yu[%d] = %-g' % (i,yu[i])
            comp = max(comp, (Ucon[i] - c[i]) * yu[i])

        # Dual feasibility
        J = kwargs.get('jac', self.jac(x))
        yloc = y.copy()
        for i in upperC:
            yloc[i] = -yloc[i]
        for i in rangeC:
            yloc[i] -= yu[i]
        w = z.copy()
        for i in upperB:
            w[i] = -w[i]
        for i in rangeB:
            w[i] -= zu[i]
        g = numpy.empty(self.n)
        J.matvec_transp(-y, g)
        for i in w.keys():
            g[i] -= w[i]

        # Include gradient of objective with appropriate multiplier if
        # checking Fritz-John conditions
        gradf = kwargs.get('grad', self.grad(x))
        if not FJ:
            g += gradf
        else:
            if objMult != 0.0: g += objMult * gradf

        dFeas = numpy.linalg.norm(g, ord=numpy.inf)

        return(dFeas, comp, bndComp, pFeas, bndFeas)


    # Check wether optimality residuals fall below
    # the prescribed thresholds.

    def AtOptimality(self, x, y, z, **kwargs):
        """
        AtOptimality(x, y, z, **kwargs)
        See OptimalityResiduals for a description of the arguments.
        Returns a tuple (res, optimal) where
        
        res      is the tuple of residuals, as returned by OptimalityResiduals,
        optimal  is True if the residuals fall below the threshold specified
                 by self.stop_d, self.stop_c and self.stop_p.
        """

        # Obtain optimality residuals
        res = self.OptimalityResiduals(x, y, z, **kwargs)

        df = res[0];              stopd = self.stop_d
        cp = max(res[1], res[2]); stopc = self.stop_c
        pf = max(res[3], res[4]); stopp = self.stop_p
        
        return (res, (df <= stopd) and (cp <= stopc) and (pf <= stopp))

    # The following methods mirror the module
    # functions defined in _amplpy.c.

    def obj(self, x):
        """
        Evaluate objective function value at x.
        Returns a floating-point number.
        """
        self.feval += 1
        return _amplpy.eval_obj(x)

    def grad(self, x):
        """
        Evaluate objective gradient at x.
        Returns a numpy array.
        """
        self.geval += 1
        return _amplpy.grad_obj(x)

    def sgrad(self, x):
        """
        Evaluate sparse objective gradient at x.
        Returns a sparse vector.
        """
        try:
            sg_dict = _amplpy.eval_sgrad(x)
        except:
            raise RunTimeError, "Failed to fetch sparse gradient of objective"

        self.geval += 1
        sg = sv.SparseVector(self.n, sg_dict)
        return sg

    def cost(self):
        """
        Evaluate sparse cost vector.
        Useful when problem is a linear program.
        Return a sparse vector.
        """
        try:
            sc_dict = _amplpy.eval_cost()
        except:
            raise RunTimeError, "Failed to fetch sparse cost vector"

        sc = sv.SparseVector(self.n, sc_dict)
        return sc

    def cons(self, x):
        """
        Evaluate vector of constraints at x.
        Returns a numpy array.
        """
        try:
            c = _amplpy.eval_cons(x)
        except:
            print ' Offending argument : ', x
            c = self.Infinity * numpy.ones(self.m)
        self.ceval += self.m
        return c

    def consPos(self, x):
        """
        Convenience function to return the vector of constraints when the latter
        are reformulated as equalities and 'greater than' inequalities only. In
        other words, the constraints of the problem are reformulated as

            ci(x) - ai  = 0  for i in equalC
            ci(x) - Li >= 0  for i in lowerC
            ci(x) - Li >= 0  for i in rangeC
            Ui - ci(x) >= 0  for i in upperC
            Ui - ci(x) >= 0  for i in rangeC.

        The constraints appear in natural order, except for the fact that the
        'upper side' of range constraints is appended to the list.
        """
        m = self.m
        equalC = self.equalC
        lowerC = self.lowerC
        upperC = self.upperC
        rangeC = self.rangeC ; nrangeC = self.nrangeC

        c = numpy.empty(m + nrangeC)
        c[:m] = self.cons(x)
        c[m:] = c[rangeC]

        c[equalC] -= self.Lcon[equalC]
        c[lowerC] -= self.Lcon[lowerC]
        c[upperC] -= self.Ucon[upperC] ; c[upperC] *= -1
        c[rangeC] -= self.Lcon[rangeC]
        c[m:]     -= self.Ucon[rangeC] ; c[m:] *= -1

        return c

    def icons(self, i, x):
        """
        Evaluate value of i-th constraint at x.
        Returns a floating-point number.
        """
        self.ceval += 1
        return _amplpy.eval_ci(i, x)

    def igrad(self, i, x):
        """
        Evaluate dense gradient of i-th constraint at x.
        Returns a numpy array.
        """
        self.Jeval += 1
        return _amplpy.eval_gi(i, x)

    def sigrad(self, i, x):
        """
        Evaluate sparse gradient of i-th constraint at x.
        Returns a sparse vector representing the sparse gradient
        in coordinate format.
        """
        try:
            sci_dict = _amplpy.eval_sgi(i, x)
        except:
            raise RunTimeError, "Failed to fetch sparse constraint gradient"

        self.Jeval += 1
        sci = sv.SparseVector(self.n, sci_dict)
        return sci

    def irow(self, i):
        """
        Evaluate sparse gradient of the linear part of the
        i-th constraint. Useful to obtain constraint rows
        when problem is a linear programming problem.
        """
        try:
            sri_dict = _amplpy.eval_row(i)
        except:
            raise RunTimeError, "Failed to fetch sparse row"

        sri = sv.SparseVector(self.n, sri_dict)
        return sri
    
    def A(self, *args):
        """
        Evaluate sparse Jacobian of the linear part of the
        constraints. Useful to obtain constraint matrix
        when problem is a linear programming problem.
        """
        if len(args) == 1:
            if type(args[0]).__name__ == 'll_mat':
                return _amplpy.eval_A(args[0])
            else:
                return None
        return _amplpy.eval_A()

    def jac(self, x, *args):
        """
        Evaluate sparse Jacobian of constraints at x.
        Returns a sparse matrix in format self.mformat
        (0 = linked list, 1 = coordinate format).
        """
        if len(args) > 0:
            if type(args[0]).__name__ == 'll_mat':
                self.Jeval += 1
                return _amplpy.eval_J(x, self.mformat, args[0])
            else:
                return None
        else:
            self.Jeval += 1
            return _amplpy.eval_J(x, self.mformat)

    def jacPos(self, x):
        """
        Convenience function to evaluate the Jacobian matrix of the constraints
        reformulated as

            ci(x) = ai       for i in equalC
            ci(x) - Li >= 0  for i in lowerC
            ci(x) - Li >= 0  for i in rangeC
            Ui - ci(x) >= 0  for i in upperC
            Ui - ci(x) >= 0  for i in rangeC.

        The gradients of the general constraints appear in
        'natural' order, i.e., in the order in which they appear in the problem.
        The gradients of range constraints appear in two places: first in the
        'natural' location and again after all other general constraints, with a
        flipped sign to account for the upper bound on those constraints.

        The overall Jacobian of the new constraints thus has the form

           n
        +-----+
        |  J  |   <-- general constraints (natural order)
        +-----+
        | -JR |   <-- 'upper' side of range constraints
        +-----+

        where the signs corresponding to 'upper' constraints and upper bounds
        should be flipped in the (1,1) block.
        """
        n = self.n ; m = self.m ; nrangeC = self.nrangeC    

        # Initialize sparse Jacobian
        J = sp(nrow=m + nrangeC, ncol=n, sizeHint=self.nnzj+10*nrangeC)

        # Insert contribution of general constraints
        J[:m,:n] = self.jac(x)
        J[upperC,:n] *= -1                 # Flip sign of 'upper' gradients
        J[m:,:n] = -J[rangeC,:n]           # Append 'upper' side of range const.
        return J

    def hess(self, x, z, *args):
        """
        Evaluate sparse lower triangular Hessian at (x, z).
        Returns a sparse matrix in format self.mformat
        (0 = linked list, 1 = coordinate format).
        """
        if len(args) > 0:
            if type(args[0]).__name__ == 'll_mat':
                self.Heval += 1
                return _amplpy.eval_H(x, z, self.mformat, args[0])
            else:
                return None
        else:
            self.Heval += 1
            return _amplpy.eval_H(x, z, self.mformat)

    def hprod(self, z, v):
        """
        Evaluate matrix-vector product H(x,z) * v.
        The x used is the last x at which the objective or its gradient was
        evaluated.
        Returns a numpy array.
        """
        self.Hprod += 1
        return _amplpy.H_prod(z, v)

    def islp(self):
        """
        Determines whether problem is a linear programming problem.
        """
        if _amplpy.is_lp():
            return True
        else:
            return False

    def set_x(self,x):
        """
        Set x as current value for subsequent calls to obj, grad, jac, etc. If
        several of obj, grad, Jac, ..., will be called with the same argument x,
        it may be more efficient to first call set_x(x). In AMPL, obj, grad,
        etc., normally check whether their argument has changed since the last
        call. Calling set_x() skips this check.
        """
        return _amplpy.set_x(x)

    def unset_x(self):
        """
        Reinstates the default behavior of obj, grad, Jac, etc., which is to
        check whether their argument has changed since the last call.
        """
        return _amplpy.unset_x()

#########################################################################
