# Implement the Limited-Memory BFGS Algorithm

from lbfgs_class import LbfgsUpdate
from pyswolfe import StrongWolfeLineSearch
import norms
from nlpy_timing import cputime

class LbfgsFramework:
    """
    Class LbfgsFramework provides a framework for solving unconstrained
    optimization problems by means of the limited-memory BFGS method.

    Instantiation is done by

        lbfgs = LbfgsFramework( nlp )

    where nlp is an instance of a nonlinear problem. A solution of the
    problem is obtained by called the solve member function, as in

        lbfgs.solve().

    Optional keyword arguments include

        npairs      the number of (s,y) pairs to store (default: 5)
        x0          the starting point (default: nlp.x0)
        maxiter     the maximum number of iterations (default: max(10n,1000))
        
    The linesearch used in this version is the More and Thuente linesearch,
    attempting to ensure satisfaction of the strong Wolfe conditions.
    Difficulties may occur from the accumulation of rounding errors inherent
    to this linesearch procedure. 
    $Id:$
    """
    def __init__( self, nlp, **kwargs ):

        self.nlp = nlp
        self.npairs = kwargs.get( 'npairs', 5 )
        self.iter   = 0
        self.nresets = 0
        self.converged = False
        
        self.lbfgsupdate = LbfgsUpdate( self.nlp.n, **kwargs )

        self.x = kwargs.get( 'x0', self.nlp.x0 )
        self.f = self.nlp.obj( self.x )
        self.g = self.nlp.grad( self.x )
        self.gnorm = norms.norm2( self.g )
        self.f0 = self.f
        self.g0 = self.gnorm

        # Optional arguments
        self.maxiter = kwargs.get( 'maxiter', max( 10*self.nlp.n, 1000 ))
        self.tsolve = 0.0

    def solve( self ):

        tstart = cputime()

        # Initial search direction is the steepest descent direction
        # In other words, the initial LBFGS matrix is the identity
        #d = -self.g
        msg = ''
        stoptol = self.nlp.stop_d * max( 1.0, norms.norm2( self.x ) )

        while self.gnorm > stoptol and self.iter < self.maxiter:

            print '%-5d  %-12g  %-12g  %-15s' % (self.iter, self.f, self.gnorm, msg)

            # Obtain search direction
            d = self.lbfgsupdate.matvec( self.iter, -self.g )

            # Prepare for More-Thuente linesearch
            SWLS = StrongWolfeLineSearch( self.f,
                                          self.g,
                                          d,
                                          lambda t: self.nlp.obj(self.x+t*d),
                                          lambda t: self.nlp.grad(self.x+t*d),
                                          stp = 1.0 )
            # Perform linesearch
            SWLS.search()
            msg = SWLS.message

            if SWLS.stp is None:
                print 'Error:: linesearch says: ', SWLS.message
                self.tsolve = cputime() - tstart
                return
            
            s = SWLS.stp * d
            self.x += s
            y = SWLS.g - self.g
            self.g = SWLS.g
            self.gnorm = norms.norm2( self.g )
            self.f = SWLS.f
            stoptol = self.nlp.stop_d * max( 1.0, norms.norm2( self.x ) )

            # Update inverse Hessian approximation using the most recent pair
            self.lbfgsupdate.store( self.iter, s, y )
            self.iter += 1

        self.tsolve = cputime() - tstart
        self.converged = (self.iter < self.maxiter)
            

if __name__ == '__main__':

    import amplpy
    import generic_io
    import sys

    ProblemName = generic_io.parse_cmdline( sys.argv[1:] )
    t_setup = cputime()
    nlp = amplpy.AmplModel( ProblemName )
    if nlp.nbounds > 0 or nlp.m > 0:
        sys.stderr.write( 'Problem has bounds or general constraints\n' )
        sys.stderr.write( 'Aborting.\n' )
        nlp.close()
        sys.exit(1)
    
    lbfgs = LbfgsFramework( nlp, npairs = 5, scaling = False )
    t_setup = cputime() - t_setup

    lbfgs.solve()

    # Output final statistics
    print
    print '-------------------------------'
    print 'LBFGS: End of Execution'
    print '  Problem                     : %-s' % ProblemName
    print '  Dimension                   : %-d' % nlp.n
    print '  Number of (s,y) pairs stored: %-d' % lbfgs.npairs
    print '  Converged to optimality     : %-s' % repr(lbfgs.converged)
    print '  Initial/Final Objective     : %-g/%-g' % (lbfgs.f0, lbfgs.f)
    print '  Initial/Final Gradient Norm : %-g/%-g' % (lbfgs.g0,lbfgs.gnorm)
    print '  Number of iterations        : %-d' % lbfgs.iter
    print '  Scaling                     : %-s' % repr(lbfgs.lbfgsupdate.scaling)
    print '  Setup/Solve time            : %-gs/%-gs' % (t_setup,lbfgs.tsolve)
    print '  Total time                  : %-gs' % (t_setup + lbfgs.tsolve)
    print '-------------------------------'

    #nlp.writesol( lbfgs.x, nlp.pi0, 'And the winner is' ) # Output  solution
    nlp.close()                                # Close connection with model
