# Implement the Limited-Memory BFGS Algorithm

from lbfgs_class import LbfgsUpdate
from pymswolfe import StrongWolfeLineSearch   # Modified More-Thuente linesearch
import norms
from nlpy_timing import cputime

class LbfgsFramework:
    """
    Class LbfgsFramework provides a framework for solving unconstrained
    optimization problems by means of the limited-memory BFGS method.

    Instantiation is done by

        lbfgs = LbfgsFramework( nlp )

    where nlp is an instance of a nonlinear problem. A solution of the
    problem is obtained by called the solve member function, as in

        lbfgs.solve().

    Optional keyword arguments include

        npairs      the number of (s,y) pairs to store (default: 5)
        x0          the starting point (default: nlp.x0)
        maxiter     the maximum number of iterations (default: max(10n,1000))

    Other keyword arguments will be passed to LbfgsUpdate.
        
    The linesearch used in this version is Jorge Nocedal's modified More and
    Thuente linesearch, attempting to ensure satisfaction of the strong Wolfe
    conditions. The modifications attempt to limit the effects of rounding
    error inherent to the More and Thuente linesearch.

    $Id:$
    """
    def __init__( self, nlp, **kwargs ):

        self.nlp = nlp
        self.npairs = kwargs.get( 'npairs', 5 )
        self.silent = kwargs.get( 'silent', False )
        self.iter   = 0
        self.nresets = 0
        self.converged = False
        
        self.lbfgsupdate = LbfgsUpdate( self.nlp.n, **kwargs )

        self.x = kwargs.get( 'x0', self.nlp.x0 )
        self.f = self.nlp.obj( self.x )
        self.g = self.nlp.grad( self.x )
        self.gnorm = norms.norm2( self.g )
        self.f0 = self.f
        self.g0 = self.gnorm

        # Optional arguments
        self.maxiter = kwargs.get( 'maxiter', max( 10*self.nlp.n, 1000 ))
        self.tsolve = 0.0

    def solve( self ):

        tstart = cputime()

        # Initial LBFGS matrix is the identity. In other words,
        # the initial search direction is the steepest descent direction
        stoptol = self.nlp.stop_d * max( 1.0, norms.norm2( self.x ) )

        while self.gnorm > stoptol and self.iter < self.maxiter:

            if not self.silent:
                print '%-5d  %-12g  %-12g' % (self.iter, self.f, self.gnorm)

            # Obtain search direction
            d = self.lbfgsupdate.matvec( self.iter, -self.g )

            # Prepare for modified More-Thuente linesearch
            if self.iter == 0:
                stp0 = 1.0/self.gnorm
            else:
                stp0 = 1.0
            SWLS = StrongWolfeLineSearch( self.f,
                                          self.x,
                                          self.g,
                                          d,
                                          lambda z: self.nlp.obj(z),
                                          lambda z: self.nlp.grad(z),
                                          stp = stp0 )
            # Perform linesearch
            SWLS.search()

            # SWLS.x  contains the new iterate
            # SWLS.g  contains the objective gradient at the new iterate
            # SWLS.f  contains the objective value at the new iterate
            s = SWLS.x - self.x
            self.x = SWLS.x
            y = SWLS.g - self.g
            self.g = SWLS.g
            self.gnorm = norms.norm2( self.g )
            self.f = SWLS.f
            stoptol = self.nlp.stop_d * max( 1.0, norms.norm2( self.x ) )

            # Update inverse Hessian approximation using the most recent pair
            self.lbfgsupdate.store( self.iter, s, y )
            self.iter += 1

        self.tsolve = cputime() - tstart
        self.converged = (self.iter < self.maxiter)
            

if __name__ == '__main__':

    import amplpy
    import generic_io
    import sys

    ProblemName = generic_io.parse_cmdline( sys.argv[1:] )
    t_setup = cputime()
    nlp = amplpy.AmplModel( ProblemName )
    if nlp.nbounds > 0 or nlp.m > 0:
        sys.stderr.write( 'Problem has bounds or general constraints\n' )
        sys.stderr.write( 'Aborting.\n' )
        nlp.close()
        sys.exit(1)

    nlp.stop_d = 1.0e-12
    
    lbfgs = LbfgsFramework( nlp, npairs = 5, scaling = True, silent = False )
    t_setup = cputime() - t_setup

    lbfgs.solve()

    # Output final statistics
    print
    print '-------------------------------'
    print 'LBFGS: End of Execution'
    print '  Problem                     : %-s' % ProblemName
    print '  Dimension                   : %-d' % nlp.n
    print '  Number of (s,y) pairs stored: %-d' % lbfgs.npairs
    print '  Converged to optimality     : %-s' % repr(lbfgs.converged)
    print '  Initial/Final Objective     : %-g/%-g' % (lbfgs.f0, lbfgs.f)
    print '  Initial/Final Gradient Norm : %-g/%-g' % (lbfgs.g0,lbfgs.gnorm)
    print '  Number of iterations        : %-d' % lbfgs.iter
    print '  Scaling                     : %-s' % repr(lbfgs.lbfgsupdate.scaling)
    print '  Setup/Solve time            : %-gs/%-gs' % (t_setup,lbfgs.tsolve)
    print '  Total time                  : %-gs' % (t_setup + lbfgs.tsolve)
    print '-------------------------------'

    #nlp.writesol( lbfgs.x, nlp.pi0, 'And the winner is' ) # Output  solution
    nlp.close()                                # Close connection with model
