# A pure Python/numpy implementation of the preconditioned
# conjugate gradient algorithm. Compare to the itsolvers.pcg
# solver from PySparse.

import numpy
from math import sqrt
import sys

def pcg( b, **kwargs ):
    """
    Solve the n-by-n symmetric positive-definite linear system of equations

        A x = b,

    or equivalenty, solve the convex quadratic minimization problem

        minimize  < -b, x > + 1/2 < x, Ax >

    by means of the so-called conjugate gradient method.

    Usage:

        (x,it,relres,converged) = pcg(A,b)

    The return values are

        x          the final estimate of the solution
        it         the number of iterations
        relres     the final relative residual
        converged  True if the iteration converged before the maximum
                   number of iterations was reached.

    Accepted keywords are

        A          the explicit coefficient matrix
        matvec     a function to compute matrix-vector products with A
        atol       the absolute tolerance (default: 1.0e-8)
        rtol       the relative tolerance (default: 1.0e-6)
        maxit      the maximum number of iterations (default: 2n)
        prec       a user-defined preconditioner.

    The relative residual is defined by ||Ax-b||/||b|| in Euclidian norm.

    At least one of 'A' or 'matvec' must be given.
    If the 'matvec' function is given, it takes precedence over 'A'.
    If 'A' is given, it must have a method named 'matvec' to compute
    matrix-vector products.

    In theory, the method should not take more than n iterations to converge.
    However, if the matrix A is not well conditioned, it may take much more
    or not converge at all. A preconditioner may then be speficied by means of
    the keyword 'prec'. prec should be a function with the following prototype

        y = prec(z)

    accepting an n-vector z as input and returning the n-vector y solving the
    linear system M y = z, where M is the preconditioner. It may for instance be
    used in conjunction with the precon module as follows:

       K = precon.BandedPreconditioner( A, bandwidth = 10 )
       (x, niter, relres, converged ) = pcg.pcg( b, A = A, prec = K.precon )

    $Id:$
    """
    n = len(b)

    matvec = kwargs.get( 'matvec', None )
    if matvec is None:
        A = kwargs.get( 'A', None )
        if A is None:
            raise ValueError, 'One of A or matvec must be given'
            return None
    atol = kwargs.get( 'atol', 1.0e-8 )      # Default absolute tolerance
    rtol = kwargs.get( 'rtol', 1.0e-6 )      # Default relative tolerance
    maxit = kwargs.get( 'maxit', 2*n )       # Default max # iterations
    prec = kwargs.get( 'prec', lambda v: v ) # Default preconditioner = identity
    debug = kwargs.get( 'debug', False )

    header = '%-5s  %-21s  %-8s\n' % ('Iter', 'Relres', 'Curv')
    fmt = '%-5d  %-21.15e  %-8.2e\n'

    # Initialization
    x = numpy.zeros( n, 'd' )  # x0 = 0
    r = -b                       # r = A x0 - b = -b

    y = prec( r )                # Preconditioned residual

    res = numpy.dot( r, y )
    sqrtres = sqrt(res)
    sqrtres0 = sqrtres            # Initial preconditioned residual norm
    p = -y
    k = 0

    tol = max( atol, rtol*sqrtres0 )  # Relative tolerance
    if matvec is None: Ap = numpy.empty( n, 'd' )

    if debug:
        sys.stderr.write( header )
        sys.stderr.write( '-' * len(header) + '\n' )

    while sqrtres > tol and k < maxit:

        if matvec is not None:
            Ap = matvec(p)
        else:
            A.matvec( p, Ap )

        pAp = numpy.dot( p, Ap )

        if debug: sys.stderr.write( fmt % (k, sqrtres, pAp) )

        if pAp <= 0.0:
            raise TypeError, 'Matrix must be positive definite'
            return None

        alpha = res/pAp
        x += alpha * p
        r += alpha * Ap
        y = prec( r )
        res1 = numpy.dot( r, y )

        if prec is not None and res1 < 0.0:
            raise TypeError, 'Preconditioner not positive definite'
            return None

        beta = res1/res
        res = res1
        sqrtres = sqrt(res)
        p = -y + beta * p
        k += 1

    if debug: sys.stderr.write( '-' * len(header) + '\n' )
    return( x, k, sqrtres/sqrtres0, k < maxit )

def A_prod( A, v ):
    # For simulation purposes only
    n = A.shape[0]
    Av = numpy.zeros( n, 'd' )
    A.matvec( v, Av )
    return Av
    
if __name__ == '__main__':

    from pysparse import spmatrix
    import precon
    from nlpy_timing import cputime
    import sys
    import os
    import norms

    header_s = '%8s  %6s  %5s  %6s  %8s  %7s  %5s  %6s  %6s\n'
    header_d = '%8s  %6d  %5d  %6d  %8.1e  %7.1e  %5s  %6.2f  %6.2f\n'
    sys.stdout.write( header_s % ('Matrix', 'n', 'nnz', 'iter', '||x-e||', 'relres', 'conv', 'setup', 'solve' ) )
    probList = sys.argv[1:]

    for problem in probList:

        prob = os.path.basename( problem )
        if prob[-4:] == '.mtx': prob = prob[:-4]

        t_setup = cputime()
        A = spmatrix.ll_mat_from_mtx( problem )
        n = A.shape[0]
        b = numpy.empty( n, 'd' )
        e = numpy.ones( n, 'd' )
        A.matvec( e, b )
        K = precon.BandedPreconditioner(A, bandwidth=5)
        t_setup = cputime() - t_setup

        # Solve system Ax=b. Solution is e.
        t_solve = cputime()
        (x,k,relres,converged) = pcg( b,
                                      A = A,
                                      prec = K.precon,
                                     )
        #(x,k,relres,converged) = pcg(b, matvec = lambda p: A_prod(A,p) )
        t_solve = cputime() - t_solve

        sys.stdout.write( header_d % (prob, n, A.nnz, k, norms.norm_infty(x-e), relres, repr(converged), t_setup, t_solve ))
