# Long-step primal-dual interior-point method for linear programming
# From Algorithm IPF on p.110 of Stephen J. Wright's book
# "Primal-Dual Interior-Point Methods", SIAM ed., 1997.
# The method uses the augmented system formulation. These systems
# are solved using PyMa27.
#
# D. Orban, Montreal 2004.

import amplpy
import pyma27
from pysparse import spmatrix
import norms
import numpy
import sparse_vector_class as sv
from nlpy_timing import cputime
import sys

################### Helper functions

def usage():
    sys.stderr.write( 'Use: %-s problem_name\n' % sys.argv[0] )
    sys.stderr.write( ' where problem_name represents a linear program\n' )

def SolveSystem( P, rhs, itref_threshold = 1.0e-5, nitrefmax = 5 ):
    P.solve( rhs )
    nr = norms.norm2( P.residual )
    # If residual not small, perform iterative refinement
    P.refine( rhs, tol = itref_threshold, nitref = nitrefmax )
    nr1 = norms.norm2( P.residual )
    return (P.x, nr1, P.neig)

def whereneg( d ):
    return numpy.nonzero( numpy.where( numpy.less( d, 0 ), d, 0) )[0].tolist()

def MaxStepLength( x, d ):
    """
    Returns the max step length from x to the boundary
    of the nonnegative orthant in the direction d
      alpha_max = min [ 1, min { -x[i]/d[i] | d[i] < 0 } ].
    Note that 0 < alpha_max <= 1.
    """
    dxneg = [ -x[i]/d[i] for i in whereneg( d ) ]
    dxneg.append( 1 )
    return min( dxneg )

def PDInteriorPointLPAmpl( lp, **kwargs ):
    """
    PDInteriorPointLPAmpl( lp, kwargs )
    solves the linear program described by the AmplModel object lp.
    See PDInteriorPointLP() for more details.
    """
    if not lp.islp():
        sys.stderr.write( ' Problem %-s is nonlinear\n' % probname )
        sys.stderr.write( ' Can only solve linear programs\n' )
        return None
    if lp.nrangeC + lp.nlowerC + lp.nupperC > 0:
        sys.stderr.write( ' Only want problems in standard form for now.\n' )
        return None
    return PDInteriorPointLP( lp.A(), lp.Lcon, lp.cost(), **kwargs )

def PDInteriorPointLP( A, b, c, **kwargs ):
    """
    (x, y, s, obj_value, iter, residual, solve_time, status)
     = PDInteriorPointLP( A, b, c)

    solves the linear program
        minimize     c'x
        subject to   Ax = b, x >= 0,
    where c is a sparse cost vector, A is a sparse constraint matrix and b is a
    dense right-hand side.

    The values returned are
         x..............the final iterate
         y..............the final value of the Lagrange multipliers associated
                        to Ax=b
         s..............the final value of the Lagrange multipliers associated
                        to x>=0
         obj_value......the final cost
         iter...........the total number of iterations
         residual.......the final relative residual
         solve_time.....the time to solve the LP
         status.........a string describing the exit status.

    Keyword arguments may be passed to change options. They are as follows:
         debug                True/False (default: False)
              Will output some debugging info during the run.
              
         PredictorCorrector   True/False (default: True)
              Uses Mehrotra's predictor-corrector method. If set
              to False, a classical long-step method will be used.
              In both cases, the augmented system formulation is
              used and systems are solved with MA27.

         itermax              integer (default: 100)
              Sets the maximum number of iterations.

         tolerance            float (default: 1.0e-6)
              The algorithm terminates successfully as soon as the
              relative residual of the KKT conditions is smaller
              than 'tolerance' in infinity norm.          
    """

    opt_args = { 'debug' : False,
                 'verbose' : False,
                 'PredictorCorrector' : True,
                 'itermax' : 100,
                 'tolerance' : 1.0e-6
               }

    for oarg in kwargs.keys():
        if opt_args.has_key( oarg ):
            opt_args[ oarg ] = kwargs[ oarg ]

    debug = opt_args['debug']
    verbose = opt_args['verbose']
    PredictorCorrector = opt_args['PredictorCorrector']
    itermax = opt_args['itermax']
    tolerance = opt_args['tolerance']

    m, n = A.shape

    normbc = 1 + max( norms.norm_infty(b), sv.norm_infty(c) )
    bigM = max( A.norm('inf'), normbc )

    x = 100 * bigM * numpy.ones( n, 'd' )
    s = x.copy()
    y = numpy.zeros( m, 'd' )
    ry = numpy.zeros( n, 'd' )
    rb = numpy.zeros( m, 'd' )
    rc = numpy.zeros( n, 'd' )
    ds = numpy.zeros( n, 'd' )
    rhs = numpy.zeros( n + m, 'd' )
    finished = False
    iter = 0

    format_hdr = '%-4s  %9s  %-8s  %-7s  %-4s  %-4s  %-8s'
    format1 = '%-4d  %9.2e  %-8.2e'
    format2 = '  %-7.1e  %-4.2f  %-4.2f  %-8.2e'

    header = format_hdr % ( 'Iter', 'Cost', 'Residual', 'Mu', 'AlPr', 'AlDu', 'IR Resid' )
    sys.stdout.write( header + '\n' )
    sys.stdout.write( '-' * len( header ) + '\n' )

    setup_time = cputime()

    # Initialize augmented matrix
    H = spmatrix.ll_mat_sym( n + m, n + A.nnz )
    H[n:, :n] = A
            
    # Main loop

    while not finished:

        if debug:
            sys.stderr.write( ' s = ', s )
            sys.stderr.write( ' x = ', x )

        # Compute residuals
        #  rb = Ax - b
        A.matvec( x, rb )
        rb -= b
        #  rc = Xs
        for i in range( n ): rc[i] = x[i] * s[i]
        #  ry = A^T y + s - c
        A.matvec_transp( y, ry )
        ry += s
        for k in c.keys(): ry[k] -= c[k]
    
        nrb = norms.norm_infty( rb )
        nrc = norms.norm_infty( rc )
        nry = norms.norm_infty( ry )
        if debug:
            sys.stderr.write( ' nrb, nrc, nry = ', nrb, nrc, nry )
        residual = max( nrb, nrc, nry ) / normbc
        obj_value = sv.dot( c, x )

        sys.stdout.write( format1 % (iter, obj_value, residual) )
    
        if residual <= tolerance:
            status = 'Optimal solution found'
            finished = True
        elif  iter > itermax:
            status = 'Max number of iterations reached'
            finished = True
        else:
            # Solve the linear system
            #
            # [ -S X^{-1}    A^T ]  [ dx ] = - [ ry - X^{-1} rc ]
            # [   A          0   ]  [ dy ]     [ rb             ]
            #
            # and recover  ds = -X^{-1} ( rc + S dx )
            # with rc = Xs - sigma * mu e.

            # Compute augmented matrix and factorize it
            for i in range( n ): H[i,i] = -s[i]/x[i]
            P = pyma27.PyMa27Context( H )

            # Compute mu
            mu = sum( rc )/n
            tau = max( .9995, 1.0 - mu )

            if PredictorCorrector:
                # Use Mehrotra predictor-corrector method
                # Compute affine-scaling step, i.e. with sigma = 0
                rhs[:n] = -(ry - s)
                rhs[n:] = -rb
                (step, nres, neig) = SolveSystem( P, rhs )
                # Recover dx and ds
                dx = step[:n]
                for i in range( n ): ds[i] = -s[i] * ( 1 + dx[i]/x[i] )
                # Compute primal and dual stepsizes
                alpha_p = MaxStepLength( x, dx )
                alpha_d = MaxStepLength( s, ds )
                # Estimate duality gap after affine-scaling step
                muAff = numpy.dot( x + alpha_p * dx, s + alpha_d * ds )/n
                sigma = (muAff/mu)**3

                if debug:
                    sys.stderr.write( ' alpha_pAFF, alpha_dAFF, muAFF, sigma =',
                                      (alpha_p, alpha_d, muAff, sigma) )

                # Incorporate predictor information for corrector step
                for i in range( n ): rc[i] += dx[i]*ds[i]
            else:
                # Use long-step method
                # Compute right-hand side
                sigma = min(0.1, 100*mu)

            # Store right-hand side
            rc -= sigma * mu
            for i in range( n ): ry[i] -= rc[i]/x[i]
            rhs[:n] = -ry
            rhs[n:] = -rb

            # Solve augmented system
            (step, nres, neig) = SolveSystem( P, rhs )

            # Recover step
            dx = step[:n]
            dy = step[n:]
            for i in range( n ): ds[i] = -(rc[i] + s[i]*dx[i])/x[i]

            # Compute primal and dual stepsizes
            alpha_p = tau * MaxStepLength( x, dx )
            alpha_d = tau * MaxStepLength( s, ds )

            sys.stdout.write( format2 % (mu, alpha_p, alpha_d, nres) )
            sys.stdout.write( '\n' )

            # Update iterates
            x += alpha_p * dx
            s += alpha_d * ds
            y += alpha_d * dy

            iter += 1

    solve_time = cputime() - setup_time
    sys.stdout.write( '\n' )
    sys.stdout.write( '-' * len( header ) + '\n' )
    if verbose:
        sys.stdout.write( ' Final x = ', x )
        sys.stdout.write( ' Final s = ', s )
        sys.stdout.write( ' Final y = ', y )

    return (x, y, s, obj_value, iter, residual, solve_time, status)

############################################################

if __name__ == '__main__':

    if len( sys.argv ) < 2:
        usage()
        sys.exit( -1 )

    probname = sys.argv[1]

    lp = amplpy.AmplModel( probname )

    (x, y, s, obj_value, iter, residual, solve_time, status) = PDInteriorPointLPAmpl( lp, verbose = True, PredictorCorrector = True )

    sys.stdout.write( status + '\n' )
    sys.stdout.write( ' #Iterations: %-d\n' % iter )
    sys.stdout.write( ' RelResidual: %-g\n' % residual )
    sys.stdout.write( ' Final cost : %-g\n' % obj_value )
    sys.stdout.write( ' Solve time : %-fs\n' % solve_time )

    # End
    lp.close()
