# The projected gradient method for bound-constrained problems
# D. Orban, Montreal 2005.

import numpy
from norms import norm2

class ProjectedGradient:
    """
    A general class for solving bound-constrained nonlinear
    problems of the form

        minimize     f(x)
        subject to   l <= x <= u

    where l and u are fixed n-vectors. Possibly some components
    of l may be -Infinity and some components of u may be
    +Infinity.

    The problem should be described as an instance of an NLPModel.
    Bounds of the variables are represented by the vectors
    Lvar and Uvar so that the bounds on x[i] are Lvar[i] and Uvar[i].

    $Id:$
    """
    
    def __init__( self, nlp, **kwargs ):
        self.nlp = nlp
        self.best_estimate = None
        self.best_value = None
        self.iter = 0
        self.optimal = False

        self.feval = 0
        self.geval = 0
        self.Heval = 0

        if 'maxiter' in kwargs.keys():
            self.maxiter = kwargs['maxiter']
        else:
            self.maxiter = 15000

    def project( self, x ):
        """
        Perform the projection of x onto the feasible set.

                   | l[i]   if         x[i] <  l[i]
            y[i] = | x[i]   if l[i] <= x[i] <= u[i]
                   | u[i]   if u[i] <  x[i]
        """
        n = x.shape[0]
    
        if n != self.nlp.n:
            raise ValueError, 'x has wrong dimension'
            return None

        return numpy.minimum( self.nlp.Uvar,
                                numpy.maximum( self.nlp.Lvar, x ) )

    def solve( self ):
        
        # Ensure starting point is feasible
        x     = self.project( self.nlp.x0 )
        f     = self.nlp.obj( x  )
        grad  = self.nlp.grad(x)
        d     = -grad
        iter  = 0

        xt = self.project( x + d )
        n2 = norm2( x - xt )

        while n2 > self.nlp.stop_d and iter < self.maxiter:

            iter += 1
            t = 1.0
            ft = self.nlp.obj( xt )

            # Perform pseudo-Armijo linesearch
            while ft > f - 0.001/t * n2**2:
                t  = 0.2*t
                xt = self.project( x + t*d )
                ft = self.nlp.obj( xt )
                n2 = norm2( x - xt )

            x = xt
            grad = self.nlp.grad( x )
            d = -grad
            f = ft
            xt = self.project( x + d )
            n2 = norm2( x - xt )

        self.best_estimate = x
        self.best_value = f
        self.iter = iter
        self.optimal = (self.iter < self.maxiter)
        return


if __name__ == '__main__':

    from rosenbrock import RosenbrockModel
    from numpy import array

    # Adjust starting point and
    # stopping tolerance if desired
    RosenbrockModel.x0 = array( [-5.0, -5.0], 'd' )
    RosenbrockModel.stop_d = 10.0**(-5)
    RosenbrockModel.Lvar = array( [-1.0, -1.0], 'd' )
    RosenbrockModel.Uvar = array( [0.0, 0.0], 'd' )

    # Instantiate projected gradient algorithm
    PG = ProjectedGradient( RosenbrockModel, maxiter = 10 )

    # Minimize the model
    PG.solve()

    # Output final info
    print ' Optimal solution found: ', PG.optimal
    print ' Best estimate found   : ', PG.best_estimate
    print ' Value at best estimate: ', PG.best_value
    print ' Number of iterations  : ', PG.iter

    # This example is more time-consuming
    #import amplpy
    #bearing = amplpy.AmplModel( 'bearing_10' )
    #PG = ProjectedGradient( bearing, maxiter = 3500 )
    #PG.solve()

    # Output final info
    #print ' Optimal solution found: ', PG.optimal
    ##print ' Best estimate found   : ', PG.best_estimate
    #print ' Value at best estimate: ', PG.best_value
    #print ' Number of iterations  : ', PG.iter

