#
# Main driver for Psuperb
#

import sys
import psuperb_framework as S
import trustregion as T
import norms as N
import pygltr
import numpy

debug = False
verbose = False

if len( sys.argv ) < 2:
    print ' Use: %-s <problem_name>' % sys.argv[0]
    sys.exit()

ProblemName = sys.argv[1]

nlp = S.SuperbInteriorFramework( ProblemName,
                                 maxit = 200,
                                 usePD = True,
                                 mu = 1,
                                 nuEq = 16,
                                 nuS = 2,
                                 nuT = 14 )
TR = T.TrustRegionFramework( eta1 = 1.0e-4 )

# Solve a single trust-region subproblem

def SolveSubproblem( nlp, TR, x, s, t, y, yrange2, z, zrange2, u, v ):

    Phi  = nlp.Phi( x, s, t )
    gPhi = nlp.GradPhi( x, s, t )
    HPhi = nlp.HessPhi( x, s, t, y, yrange2, z, zrange2, u, v )
    tinyStep = False

    if verbose:
        print
        print '    ----------------------------'
        #print '    GradPhi = ', gPhi
        #print '    HessPhi = ', HPhi
        print '    Solving subproblem with GLTR'
        
    G = pygltr.PyGltrContext( gPhi, radius = TR.Delta )
    G.explicit_solve( HPhi.to_csr() )
    m = G.m
    l = G.mult
    step_norm = G.snorm
    cgiter = G.niter
    nc = G.nc
    ierr = G.ierr
    step = G.step

    if verbose:
        if ierr:
            print '       Error exit from GLTR: ', ierr
        print '       Predicted reduction : ', -m
        print '       Lagrange multiplier : ', l
        print '       Trust-region radius : ', TR.Delta
        print '       Step norm           : ', step_norm
        print '       # Lanczos iterations: ', cgiter
        print '       Negative curvature  : ', nc

    if step_norm <= nlp.tiny and step_norm < TR.Delta:
        tinyStep = True

    (stepx, steps, stept) = nlp.UnpackStep( step )

    if debug:
        print '       Step in x : ', stepx
        print '       Step in s : ', steps
        print '       Step in t : ', stept
        print

    (c_trial, cR_trial) = nlp.evalC( x + stepx )
    (stepy, stepyrange2, stepz, stepzrange2, stepu, stepv) = \
        nlp.ComputePDMult( x, s, t, step, y, yrange2, z, zrange2, u, v )
    alpha = nlp.get_alpha( x, s, t, step, y, yrange2, z, zrange2, u, v, stepy, stepyrange2, stepz, stepzrange2, stepu, stepv )

    if verbose:
        print '    Primal infeasibility : ', nlp.PrimalFeasibility( x + stepx )
        print '    Feasibility threshold: ', (nlp.etaE, nlp.etaI, nlp.etaX)
        print '    Updated iterate strictly feasible : ', feasible

    new_t = nlp.AddSparseVectors( t, stept, alpha = alpha )
    Phi_trial = nlp.Phi( x + alpha*stepx, s + alpha*steps, new_t )
    rho = TR.Rho( Phi, Phi_trial, m )

    if verbose:
        print '    Actual reduction   : ', Phi - Phi_trial
        print '    rho = ', rho

    TR.UpdateRadius( rho, step_norm )
    if rho >= TR.eta1:

        # If using primal-dual multipliers, update before taking step
        if nlp.usePD:
            (y, yrange2, z, zrange2, u, v ) = \
                nlp.UpdatePDMult( x, s, t, step, y, yrange2, z, zrange2, u, v, alpha = alpha )

        # Take step
        x = x + alpha*stepx
        s = s + alpha*steps
        t = new_t
            
        # If using primal multipliers, update after taking step
        if not nlp.usePD:
            (y, yrange2, z, zrange2, u, v) = nlp.PrimalMult( x, s, t )

        # Update function values
        Phi = Phi_trial
        f = nlp.obj( x )
        c = c_trial
        cR = cR_trial

        if verbose:
            print '    Step accepted'
            print '    New value of merit function : ', Phi
            print '    New value of objective      : ', f
            print '    New trust-region radius     : ', TR.Delta
            print '    Max( s )  : ', nlp.Max( s )
            print '    Max( t )  : ', nlp.Max( new_t )

        print '    Updated x : ', x
        print '    Updated s : ', s
        print '    Updated t : ', new_t

        if debug:
            print '    Updated mults : ', (y, yrange2, z, zrange2, u, v )

        res = nlp.InnerResiduals( x, s, t, y, yrange2, z, zrange2, u, v )
        print '    Barrier residuals at this stage: ', res
        print '    Target residuals               : ', (nlp.dual_feas_stop,
                                                        nlp.CSY_compl_stop,
                                                        nlp.SU_compl_stop,
                                                        nlp.XTZ_compl_stop,
                                                        nlp.TV_compl_stop )
    else:
        if verbose: print '    Step rejected'

    print '    ----------------------------'
    return (x, s, t, y, yrange2, z, zrange2, u, v, tinyStep)


# Solve an inner iteration -- mu and nu are fixed
# Perform a series of trust-region steps until
# barrier stopping conditions are met.

def SolveInnerIteration( nlp, TR, x, s, t, y, yrange2, z, zrange2, u, v ):

    n_inner = 0
    print
    print '  Stopping tolerances: ', (nlp.dual_feas_stop, nlp.CSY_compl_stop, \
        nlp.SU_compl_stop, nlp.XTZ_compl_stop, nlp.TV_compl_stop )
    
    finished = False
    tinyStep = False
    n_it_infeasible_inner = 0

    print '   Starting inner iteration at x = ', x
    
    while not finished:

        print ' Inner iteration %5d  mu = %7.1e  nuEq = %7.1e  nuS = %7.1e  nuT = %7.1e' % ( n_inner, nlp.mu, nlp.nuEq, nlp.nuS, nlp.nuT )

        res = nlp.InnerResiduals( x, s, t, y, yrange2, z, zrange2, u, v )
        if nlp.SufficientlySmall( res ):
            print ' Inner iteration stopping conditions satisfied'
            finished = True
        elif tinyStep:
            print ' Exiting inner iteration---tiny step'
            finished = True
        else:
            (x, s, t, y, yrange2, z, zrange2, u, v, tinyStep) = \
                SolveSubproblem(nlp, TR, x, s, t, y, yrange2, z, zrange2, u, v)
            n_inner += 1
            if n_inner >= nlp.max_inner:
                print ' Skipping to next outer iteration'
                finished = True

    # End while

    print '   Ending inner iteration at x = ', x

    print ' #Inner iterations: ', n_inner
    TR.ResetRadius()
    return( x, s, t, y, yrange2, z, zrange2, u, v, n_inner, tinyStep )
    

# Solve the outer iteration -- loop on mu and nu

def SolveOuterIteration( nlp, TR ):

    x = nlp.x0
    t = nlp.t
    s = nlp.s
    f = nlp.obj( x )
    g = nlp.grad( x )
    (c, cR) = nlp.evalC( x )
    Phi = nlp.merit0

    # Set initial trust-region radius
    TR.Delta = max( 0.1 * numpy.linalg.norm( g, ord=2 ), 1.0 )

    n_outer = 0
    n_inner_total = 0
    n_it_infeasible_total = 0

    print ' ==================='
    print ' Initial information'
    print
    print ' lower bounds : ', nlp.lowerB
    print ' upper bounds : ', nlp.upperB
    print ' 2side bounds : ', nlp.rangeB
    print
    print ' lower nonlinear : ', nlp.lowerC
    print ' upper nonlinear : ', nlp.upperC
    print ' range nonlinear : ', nlp.rangeC
    print ' equal nonlinear : ', nlp.equalC
    print

    (y, yrange2, z, zrange2, u, v ) = nlp.PrimalMult( x, s, t )
    #if nlp.usePD:
    #    mult_factor = 1.0
    #    y = mult_factor * numpy.ones( nlp.nConst, 'd' )
    #    yrange2 = {}
    #    for i in nlp.Range:
    #        yrange2[i] = mult_factor
    #    z = {}
    #    zrange2 = {}
    #    v = {}
    #    for i in nlp.Bounds:
    #        z[i] = mult_factor
    #        v[i] = mult_factor
    #    for i in nlp.rangeB:
    #        zrange2[i] = mult_factor
    #    u = mult_factor * numpy.ones( nlp.nConst, 'd' )
    #else:
    #    (y, yrange2, z, zrange2, u, v ) = nlp.PrimalMult( x, s, t )

    ys = nlp.ShiftedMult( y, yrange2 )
    zs = nlp.ShiftedZeta( z, zrange2 )
    gPhi = nlp.GradPhi( x, s, t )
    HPhi = nlp.HessPhi( x, s, t, y, yrange2, z, zrange2, u, v )
    res = nlp.InnerResiduals( x, s, t, y, yrange2, z, zrange2, u, v )
    
    print ' x0 = ', x

    if debug:
        print ' x0    = ', x
        print ' t0    = ', t
        print ' s0    = ', s
        print ' c(x0) = ', c
        print ' cR(x0)= ', cR
        print ' primal mults : ', (y, yrange2, z, zrange2, u, v )
        print ' shifted y = ', ys
        print ' shifted z = ', zs
        print ' GradPhi0  = ', gPhi
        print ' HessPhi0  = ', HPhi
        print ' J(x0) = ', nlp.jac( x )
        
    print ' f(x0) = ', f
    print ' PhiB0 = ', Phi
    print ' Current iterate is strictly feasible : ', nlp.CheckFeasible( c, cR, x, s, t )
    print ' Initial interior optimality residuals: ', res
    print ' ==================='

    # Could do this implicitly in AtOptimality?
    y_nlp = y.copy()
    for i in nlp.Eq:
        y_nlp[i] -= nlp.nuEq

    (nlp_res, optimal) = nlp.AtOptimality( x, y_nlp, z, pi_range = yrange2, z_range = zrange2 )
    print ' Initial NLP optimality residuals: ', nlp_res
    print ' ==================='

    # Main loop
    while not optimal and n_outer <= nlp.maxit and nlp.mu >= nlp.mu_min:

        print
        print ' ++++++++++++++++++++++++'
        print ' + Outer Iteration %-4d +' % n_outer
        print ' ++++++++++++++++++++++++'
        print

        # Solve current inner iteration
        (x, s, t, y, yrange2, z, zrange2, u, v, n_inner, tinyStep ) = \
            SolveInnerIteration( nlp, TR, x, s, t, y, yrange2, z, zrange2, u, v )

        # Update inner iteration counter
        n_inner_total += n_inner

        # Compute KKT multipliers
        # Could do this implicitly in AtOptimality?
        y_nlp = y.copy()
        for i in nlp.Eq:
            y_nlp[i] -= nlp.nuEq

        # Update outer iteration counter
        n_outer += 1

        # If max number of outer iterations has been reached, print message
        if n_outer > nlp.maxit:
            print
            print '  Maximum number of iterations reached. Terminating.'

        # Update optimality residuals
        (nlp_res, optimal) = nlp.AtOptimality( x, y_nlp, z, pi_range = yrange2, z_range = zrange2 )

        # Update penalty and barrier parameters if necessary
        nlp.UpdateMu()

        # See if mu falls below minimal threshold
        if nlp.mu < nlp.mu_min:
            print
            print '  Smallest mu has been reached. Terminating.'
        else:
            nlp.UpdateNu( x )

    # End while

    print
    if optimal:
        print 'Outer Iteration termination criterion satisfied'

    if debug:
        print 'x = ', x

    print
    header = ( 'Name', 'Final f', 'Optimal', 'Infeasibility', 'DualFeasibility', 'Complementarity', 'Outer', 'Inner' )
    print '%-12s  %-15s  %-7s  %-15s  %-15s  %-15s  %-5s  %-5s' % header
    print '%-12s  %-15g  %-7s  %-15g  %-15g  %-15g  %-5d  %-5d' % \
        ( ProblemName, nlp.obj(x), optimal, max( nlp_res[3], nlp_res[4] ), \
        nlp_res[0], max( nlp_res[1], nlp_res[2] ), n_outer-1, n_inner_total - n_it_infeasible_total )
    print
    print ' x = ', x
    print ' y = ', y
    print ' z = ', z
    print ' u = ', u
    print ' v = ', v
    print ' s = ', s
    print ' t = ', t
    print
    print ' Max s = ', nlp.Max( s )
    print ' Max t = ', nlp.MaxD( t )
    print ' Max y = ', max( nlp.InfinityNorm( y_nlp ), nlp.MaxD( yrange2 ) )
    print ' Max z = ', max( nlp.MaxD( z ), nlp.MaxD( zrange2 ) )
    return
        

# Solve
SolveOuterIteration( nlp, TR )

# Terminate
nlp.close()
