###############################
#                             #
# Class definition for Superb #
#                             #
###############################

import numpy
from pysparse import spmatrix
import math
import sys
from amplpy import AmplModel

# Class SuperbInteriorFramework inherits from AmplModel

class SuperbInteriorFramework( AmplModel ):

    """
    ipm = SuperbInteriorFramework( problem )
    Instantiate an InteriorFramework class embedding the nonlinear
    program whose name is 'problem'. As part of the embedding, the
    bound constraints are reformulated as follows
    
        xL <= x <= xU      becomes       x  - xL >= 0, and
                                         xU - x  >= 0
        xL <= x            becomes       x  - xL >= 0
        x  <= xU           becomes       xU - x  >= 0.

    The counters nrangeB, nlowerB, nupperB countain the number of bounds
    of the first, second and third kind respectively. To those are
    appended the number of fixed and free variables: nfixedB and nfreeB.
    After reformulations, there are
        nBounds = nrangeB + nlowerB + nupperB
    'effective' bound constraints.

    Similarly, constraints are reformulated as follows

        cL <= c(x)         becomes       c(x) - cL   >= 0
              c(x) <= cU   becomes       cU   - c(x) >= 0.
        cL <= c(x) <= cU   becomes       both of the above

    The counters nrangeC, nlowerC and nupperC contain the number of
    constraints of the first, second and third kind respectively. To those
    are appended the number of equality and 'free' constraints (if any):
    nequalC and nfreeC. The effective number of inequality constraints
    after reformulation s thus
        nIneq = nlowerC + nupperC.
    """
    
    # ========================================================================

    # Constructor
    def __init__( self, ProblemName, nuEq = 1.0, nuS = 1.0, nuT = 1.0,
                  mu = 5.0, maxit = 1000, max_inner = 200, usePD = True,
                  mu_min = 1.0e-10 ):

        AmplModel.__init__( self, ProblemName )

        # Classify constraints according to their type
        self.Ineq    = self.lowerC + self.upperC
        self.Eq      = self.equalC
        self.Range   = self.rangeC

        # Same for bounds -- fixed variables were removed by Ampl
        self.Bounds  = self.lowerB + self.upperB + self.rangeB
        
        # Maintain counters for effective number of constraints and bounds
        self.nBounds  = self.nrangeB + self.nlowerB + self.nupperB
        self.nBounds2 = self.nrangeB + self.nBounds
        self.nIneq    = self.nlowerC + self.nupperC
        self.nEq      = self.nequalC
        self.nRange   = self.nrangeC

        # Push the second part of range constraints at the end
        self. nConst = self.nEq + self.nIneq + self.nRange
        self.Range2  = range( self.nConst, self.nConst + self.nRange )

        # Indicate whether or not to use primal-dual multipliers
        self.usePD = usePD
        self.project_mult = True

        # Initial penalty and barrier parameters
        self.nuEq = nuEq     # Initial penalty parameter for equalities
        self.nuS  = nuS      # Initial penalty parameter for elastics
        self.nuT  = nuT      # Initial penalty parameter for bound-elastics
        self.tau1 = 2.0      # Factor of nu in updating rule
        self.tau2 = 1.0      # Constant added to nu in updating rule
        self.mu = mu         # Initial barrier parameter
        self.mufactor = 0.2  # Factor of mu in updating rule
        self.mu_min = mu_min # Smallest mu
        self.ftb = 0.95      # Fraction-to-the-boundary rule constant
        self.etaE = min( self.mu, 0.2 ) # Threshold to increase nu
        self.etaI = min( self.mu, 0.2 ) # Threshold to increase nu
        self.etaX = min( self.mu, 0.2 ) # Threshold to increase nu

        self.maxit = maxit         # Maximum number of outer iterations
        self.max_inner = max_inner # Maximum number of inner iterations
        self.dual_feas_stop = 1.1 * min( self.mu**1.0001, self.mu )
        self.CSY_compl_stop = 1.1 * self.mu
        self.SU_compl_stop  = 1.1 * self.mu
        self.XTZ_compl_stop = 1.1 * self.mu
        self.TV_compl_stop  = 1.1 * self.mu

        # Constants for the multipliers update
        self.kappaL = 0.5
        self.kappaU = 1.0e5
        
        # Constant which determines initial elastics: s = max(0,-c) + ethresh
        self.ethresh = 1.1 
        self.tiny = 1.0e-8

        # Set initial elastics so (x0, s0) is strictly feasible
        # Note that a single elastic suffices for a range constraint
        self.s = numpy.zeros( self.nConst, 'd' )
        (self.c, self.cR) = self.evalC( self.x0 )
        for i in self.Eq + self.Ineq:
            self.s[i] = max( 0.0, -self.c[i] ) + self.ethresh
        for i in self.Range:
            self.s[i] = max( 0.0, - self.cR[i],
                             - self.c[i], ) + self.ethresh

        # Set initial elastics for the bound constraints so (x0, t0)
        # strictly satisfies the bounds. Single elastic for 2-sided bounds.

        self.t = {}
        for i in self.lowerB:
            self.t[i] = max(0.0, self.Lvar[i] - self.x0[i]) + self.ethresh
        for i in  self.upperB:
            self.t[i] = max(0.0, self.x0[i] - self.Uvar[i]) + self.ethresh
        for i in self.rangeB:
            self.t[i] = max( 0.0, self.Lvar[i] - self.x0[i],
                             self.x0[i] - self.Uvar[i] ) + self.ethresh

        # Evaluate initial penalty function value
        # p(x,s,nu) = f(x) + nu ( cE(x) + 2 sE )^T e + nu sI^T e
        self.f0 = self.obj( self.x0 )
        self.p0 = self.evalP( self.x0, self.s, self.t )
        self.merit0 = self.Phi( self.x0, self.s, self.t )

        # Obtain approximate value of machine epsilon
        self.eps = self._Epsilon()

    # ========================================================================
            
    # Destructor: inherited from AmplModel

    # ========================================================================
    
    #####################
     #                 #
    ### Class methods ###
     #                 #
    #####################

    # ========================================================================

    # Reformulate constraints.
    # The "upper" side of a range constraint is pushed at the end of c;
    # If i is in Range, the index of the "upper" side is given by
    #    nConst + Range.index( i )
    def evalC( self, x ):
        c = self.cons( x )
        cR = {}
        for i in self.Eq + self.lowerC:
            c[i] -= self.Lcon[i]
        for i in self.upperC:
            c[i] = self.Ucon[i] - c[i]
        for i in self.Range:
            cR[i] = self.Ucon[i] - c[i]
            c[i] -= self.Lcon[i]
        return (c, cR)

    # ========================================================================

    # Compute c(x) + s
    def CplusS( self, c, cR, s ):
        cps = c + s
        cpsR = {}
        for i in self.Range:
            cpsR[i] = cR[i] + s[i]
        return( cps, cpsR )

    # ========================================================================

    def evalP( self, x, s, t ):
        p = self.obj( x )
        (c, cR) = self.evalC( x )
        # Add contribution from equality constraints
        for i in self.Eq:
            p += self.nuEq * ( c[i] + 2 * s[i] )
        # Add contribution from inequality and range constraints
        for i in self.Ineq + self.Range:
            p += self.nuS * s[i]
        # Add contribution from bound constraints
        for i in self.lowerB + self.upperB + self.rangeB:
            p += self.nuT * t[i]
        return p

    # ========================================================================

    # Reformulated constraints Jacobian
    # Do not reformulate range constraints --- they will be treated in place
    def evalJ( self, x ):
        return self.jac( x )

    # ========================================================================

    # Obtain primal multipliers y, z, u and v from x, s and t.
    def PrimalMult( self, x, s, t ):
        y = numpy.zeros( self.nConst, 'd' )   # General constraints on x
        z = {}                                  # Bound constraints on x
        u = numpy.zeros( self.nConst, 'd' )   # Bound constraints on s
        v = {}                                  # Bound constraints on t

        yrange2 = {}
        zrange2 = {}

        (c, cR) = self.evalC( x )
        (cps, cRps) = self.CplusS( c, cR, s )

        # Fill in u and v
        for i in range( self.nConst ):
            u[i] = self.mu * 1.0/( s[i] )
        for i in self.Bounds:
            v[i] = self.mu * 1.0/( t[i] )

        # Fill in y and z
        for i in self.Eq + self.Ineq:
            y[i] = self.mu * 1.0/cps[i]
        for i in self.Range:
            y[i] = self.mu * 1.0/cps[i]
            yrange2[i] = self.mu * 1.0/cRps[i]

        for i in self.lowerB:
            z[i] = self.mu * 1.0/( x[i] - self.Lvar[i] + t[i] )
        for i in self.upperB:
            z[i] = self.mu * 1.0/( self.Uvar[i] - x[i] + t[i] )
        for i in self.rangeB:
            z[i] = self.mu * 1.0/( x[i] - self.Lvar[i] + t[i] )
            zrange2[ i ] = self.mu * 1.0/( self.Uvar[i] - x[i] + t[i] )

        return( y, yrange2, z, zrange2, u, v )

    # ========================================================================

    # Obtain the shifted multipliers y - nu eE

    def ShiftedMult( self, y, yrange2 ):
        shifted_y = numpy.zeros( self.nConst, 'd' )
        for i in self.lowerC:
            shifted_y[i] = y[i]
        for i in self.upperC:
            shifted_y[i] = -y[i]
        for i in self.Eq:
            shifted_y[i] = y[i] - self.nuEq
        for i in self.Range:
            shifted_y[i] = y[i] - yrange2[i]
        return shifted_y

    # ========================================================================

    # Obtain the corresponding multipliers for the bounds
    def ShiftedZeta( self, z, zrange2 ):
        shifted_z = {}
        for i in self.lowerB:
            shifted_z[i] = z[i]
        for i in self.upperB:
            shifted_z[i] = -z[i]
        for i in self.rangeB:
            shifted_z[i] = z[i] - zrange2[i]
        return shifted_z

    # ========================================================================
    
    # Evaluate the barrier function and its derivatives
    
    # phi(x, mu) = p(x,s,nu) - mu * Sum log( ci(x) + si )
    #                        - mu * Sum log( si )
    #                        - mu * Sum log( bi )
    def Phi( self, x, s, t ):
        
        phi = self.evalP( x, s, t )
        (c, cR) = self.evalC( x )
        bar = 0.0
        
        for i in (self.Ineq + self.Eq):
            bar += math.log( c[i] + s[i] ) + math.log( s[i] )
        for i in self.Range:
            bar += math.log( c[i] + s[i] ) + \
                   math.log( cR[i] + s[i] ) + \
                   math.log( s[i] )
        
        for i in self.lowerB:
            bar += math.log( x[i] - self.Lvar[i] + t[i] ) + math.log( t[i] )
        for i in  self.upperB:
            bar += math.log( self.Uvar[i] - x[i] + t[i] ) + math.log( t[i] )
        for i in self.rangeB:
            bar += math.log( x[i] - self.Lvar[i] + t[i] ) + \
                   math.log( self.Uvar[i] - x[i] + t[i] ) + math.log( t[i] )

        phi -= self.mu * bar
        return phi

    # ========================================================================

    #                   [   g - J^T( y - nu eE )   ]    where y, zeta, u and v
    # grad phi(x, mu) = | nu e - ( y - nu eE ) - u |    are given by the
    #                   [     nu e - zeta - v      ]    primal estimates.
    def GradPhi( self, x, s, t ):

        # Assemble x-part of gradient
        g = self.grad( x )
        J = self.evalJ( x )
        gPhix = numpy.zeros( self.n, 'd' )
        
        # Obtain primal multipliers
        (y, yrange2, z, zrange2, u, v) = self.PrimalMult( x, s, t )

        # Obtain shifted multipliers
        shifted_y = self.ShiftedMult( y, yrange2 )
        shifted_z = self.ShiftedZeta( z, zrange2 )   # ! This is a dictionnary
        J.matvec_transp( -shifted_y, gPhix )
        gPhix += g
        for i in self.Bounds:
            gPhix[i] -= shifted_z[i]
        
        # Assemble s-part of gradient
        gPhis = self.nuS * numpy.ones( self.nConst, 'd' ) - y - u
        for i in self.Eq:
            gPhis[i] += ( 2 * self.nuEq - self.nuS )
        for i in self.Range:
            gPhis[i] -= yrange2[i]

        # Assemble t-part of gradient
        gPhit = self.nuT * numpy.ones( self.nBounds, 'd' )
        k = z.keys()
        k.sort()
        j = 0
        for i in k:
            if i in self.lowerB + self.upperB:
                gPhit[j] -= z[i] + v[i]
                j += 1
            if i in self.rangeB:
                gPhit[j] -= z[i] + zrange2[i] + v[i]
                j += 1

        # Assemble gradient
        gPhi = numpy.concatenate( (gPhix, gPhis, gPhit) )
        return gPhi

    # ========================================================================

    #                   [   g - J^T( y - nu eE )   ]    where y, zeta, u and v
    # grad phi(x, mu) = | nu e - ( y - nu eE ) - u |    are given by the
    #                   [     nu e - zeta - v      ]    primal estimates.
    def GradPhiPD( self, x, s, t, y, yrange2, z, zrange2, u, v ):

        # Assemble x-part of gradient
        g = self.grad( x )
        J = self.evalJ( x )
        gPhix = numpy.zeros( self.n, 'd' )
		
		# Obtain shifted multipliers
        shifted_y = self.ShiftedMult( y, yrange2 )
        shifted_z = self.ShiftedZeta( z, zrange2 )   # ! This is a dictionnary
	
        J.matvec_transp( -shifted_y, gPhix )
        gPhix += g
        for i in self.Bounds:
            gPhix[i] -= shifted_z[i]
     
        # Assemble s-part of gradient
        gPhis = self.nuS * numpy.ones( self.nConst, 'd' ) - y - u
        for i in self.Eq:
            gPhis[i] += ( 2*self.nuEq - self.nuS )
        for i in self.Range:
            gPhis[i] -= yrange2[i]

        # Assemble t-part of gradient
        gPhit = self.nuT * numpy.ones( self.nBounds, 'd' )
        k = z.keys()
        k.sort()
        j = 0
        for i in k:
            if i in self.lowerB + self.upperB:
                gPhit[j] -= z[i] + v[i]			
                j += 1
            if i in self.rangeB:
				gPhit[j] -= z[i] + zrange2[i] + v[i] 
				j += 1
              
        # Assemble gradient
        gPhi = numpy.concatenate( (gPhix, gPhis, gPhit) )
        return gPhi

    # ========================================================================
    
    # Evaluate the primal-dual Hessian of the barrier
    def HessPhi( self, x, s, t, yPD, yPDrange2, zPD, zPDrange2, uPD, vPD ):

        # Assemble the lower triangle of H block by block
        (c, cR) = self.evalC( x )
        J = self.evalJ( x )

        # Use the primal-dual estimates given in argument
        # Obtain shifted multipliers
        shifted_y = self.ShiftedMult( yPD, yPDrange2 )
        shifted_z = self.ShiftedZeta( zPD, zPDrange2 )

        # (2,1)-block = ThetaTilde J
        
        #H21 = J.copy()
        H21 = spmatrix.ll_mat( J.shape[0], J.shape[1], J.nnz )
        theta = numpy.zeros( self.nConst, 'd' )
        for i in self.Eq + self.lowerC:
            theta[i] = yPD[i] * 1.0/( c[i] + s[i] )
            #H21[i, : ] = J[i,:].times_scalar( theta[i] )
            # This is very inefficient; re-introduce times_scalar() !
            H21i = J[i,:]
            H21i.scale( theta[i] )
            H21[i,:] = H21i
        for i in self.upperC:
            theta[i] = - yPD[i] * 1.0/( c[i] + s[i] )
            #H21[i, : ] = J[i,:].times_scalar( theta[i] )
            H21i = J[i,:]
            H21i.scale( theta[i] )
            H21[i,:] = H21i
        for i in self.Range:
            theta[i] = yPD[i] * 1.0/( c[i] + s[i] ) - \
                       yPDrange2[i] * 1.0/( cR[i] + s[i] )
            #H21[i, : ] = J[i,:].times_scalar( theta[i] )
            H21i = J[i,:]
            H21i.scale( theta[i] )
            H21[i,:] = H21i

        #print '  thetaTilde   = ', theta
        #print '  ThetaTilde J = ', H21

        # (2,2)-block = U S^{-1} + ThetaTilde
        H22 = spmatrix.ll_mat_sym( self.nConst, self.nConst )
        #for i in self.Eq + self.Ineq + self.Range:
        #    H22[i,i] = uPD[i] / s[i] + theta[i]
        for i in self.Eq + self.lowerC:
            H22[i,i] = uPD[i] / s[i] + theta[i]
        for i in self.upperC:
            H22[i,i] = uPD[i] / s[i] - theta[i]
        for i in self.Range:
            H22[i,i] = uPD[i] / s[i] + \
                       yPD[i] * 1.0/( c[i] + s[i] ) + \
                       yPDrange2[i] * 1.0/( cR[i] + s[i] )

        #print '  U S^{-1} + Theta = ', H22

        # (3,1)-block = Xi E
        xi = {}
        H31 = spmatrix.ll_mat( self.nBounds, self.n, self.nBounds )
        for i in self.lowerB:
            xi[i] = zPD[i] * 1.0/( x[i] - self.Lvar[i] + t[i] )
        for i in self.upperB:
            xi[i] = - zPD[i] * 1.0/( self.Uvar[i] - x[i] + t[i] )
        for i in self.rangeB:
            xi[i] = zPD[i] * 1.0/( x[i] - self.Lvar[i] + t[i] ) - \
                    zPDrange2[i] * 1.0/( self.Uvar[i] - x[i] + t[i] )

        row = 0
        for j in range( self.n ):
            if j in self.lowerB + self.upperB + self.rangeB:
                H31[row, j] = xi[j]
                row +=1

        #print '  Xi E = ', H31

        # Include contribution of Jacobian in (1,1)-block
        H11tmp = spmatrix.ll_mat( J.shape[0], J.shape[1], J.nnz )
        for i in self.Eq + self.lowerC:
            #H11tmp[i, : ] = J[i,:].times_scalar( theta[i] )
            H11i = J[i,:]
            H11i.scale( theta[i] )
            H11tmp[i,:] = H11i
        for i in self.upperC:
            #H11tmp[i, : ] = J[i,:].times_scalar( -theta[i] )
            H11i = J[i,:]
            H11i.scale( -theta[i] )
            H11tmp[i,:] = H11i
        for i in self.Range:
            thetatmp = yPD[i] * 1.0/( c[i] + s[i] ) + \
                       yPDrange2[i] * 1.0/( cR[i] + s[i] )
            #H11tmp[i, : ] = J[i,:].times_scalar( thetatmp )
            H11i = J[i,:]
            H11i.scale( thetatmp )
            H11tmp[i,:] = H11i
        H11 = spmatrix.dot( J, H11tmp )
        #print '   H11tmp = ', H11
        H11.shift( 1.0, self.hess( x, -shifted_y ) )
        #for i in self.lowerB + self.upperB + self.rangeB:
        #    H11[i,i] += xi[i]
        for i in self.lowerB:
            H11[i,i] += xi[i]
        for i in self.upperB:
            H11[i,i] += -xi[i]
        for i in self.rangeB:
            H11[i,i] += zPD[i] * 1.0/( x[i] - self.Lvar[i] + t[i] ) + \
                        zPDrange2[i] * 1.0/( self.Uvar[i] - x[i] + t[i] )

        #print '  Grad Lagr   = ', self.hess( x, -shifted_y )
        #print '  (1,1) block = ', H11

        # (3,3)-block: V T^{-1} + XiTilde
        H33 = spmatrix.ll_mat_sym( self.nBounds, self.nBounds )
        elem = 0
        for i in range( self.n ):
            if i in self.lowerB:
                H33[elem,elem] = vPD[i] / t[i] + xi[i]
                elem += 1
            if i in self.upperB:
                H33[elem,elem] = vPD[i] / t[i] - xi[i]
                elem += 1
            if i in self.rangeB:
                j = self.rangeB.index(i)
                H33[elem,elem] = vPD[i] / t[i] + \
                                 zPD[i] * 1.0/( x[i] - self.Lvar[i] + t[i] ) + \
                                 zPDrange2[i] * 1.0/(self.Uvar[i] - x[i] + t[i])
                elem += 1

        #print '  V T^{-1} + XiTilde = ', H33

        H = spmatrix.ll_mat_sym( self.n + self.nConst + self.nBounds,
                                 H11.nnz + H21.nnz + self.nConst + self.nBounds)


        #      [ H11         ]
        #  H = | H21 H22     |
        #      [ H31  0  H33 ]

        for i in range( self.n ):
            for j in range( i+1 ):
                H[i,j] = H11[i,j]
        block2 = self.n+self.nConst
        H[self.n:block2, :self.n] = H21
        H[block2:, :self.n] = H31
        H[self.n:block2,self.n:block2] = H22
        H[block2:,block2:] = H33
        return H

    # ========================================================================

    # Evaluate matrix-vector product with the primal-dual Hessian of the barrier
    def HessPhiProd( self, x, s, t, yPD, yPDrange2, zPD, zPDrange2, uPD, vPD, vec ):
        # Evaluate constraints and Jacobian
        (c, cR) = self.evalC( x )
        J = self.evalJ( x )
        Hv = numpy.zeros( self.n + self.nConst + self.nBounds, 'd' )
        vecx = vec[:self.n]
        vecs = vec[self.n:self.n+self.nConst]
        vect = vec[self.n+self.nConst:]

        # Compute shifted multipliers
        shifted_y = self.ShiftedMult( yPD, yPDrange2 )

        # Form thetaTilde
        thetaTilde = numpy.zeros( self.nConst, 'd' )
        for i in self.Eq + self.lowerC:
            thetaTilde[i] = yPD[i] * 1.0/( c[i] + s[i] )
        for i in self.upperC:
            thetaTilde[i] = -yPD[i] * 1.0/( c[i] + s[i] )
        for i in self.Range:
            thetaTilde[i] = yPD[i] * 1.0/( c[i] + s[i] ) - \
                            yPDrange2[i] * 1.0/( cR[i] + s[i] )

        # Form xi
        xi = {}
        for i in self.lowerB:
            xi[i] = zPD[i] * 1.0/( x[i] - self.Lvar[i] + t[i] )
        for i in self.upperB:
            xi[i] = zPD[i] * 1.0/( self.Uvar[i] - x[i] + t[i] )
        for i in self.rangeB:
            xi[i] = zPD[i] * 1.0/(  x[i] - self.Lvar[i] + t[i] ) + \
                    zPDrange2[i] * 1.0/( self.Uvar[i] - x[i] + t[i] )

        Hvx = numpy.zeros( self.n, 'd' )
        tmp = numpy.zeros( self.nConst, 'd' )
        Jdx = numpy.zeros( self.nConst, 'd' )
        J.matvec( vecx, Jdx )                        # Jdx = J dx
        for i in self.Eq + self.lowerC:
            tmp[i] = ( Jdx[i] + vecs[i] ) * thetaTilde[i]
        for i in self.upperC:
            tmp[i] = ( - Jdx[i] + vecs[i] ) * thetaTilde[i]
        for i in self.Range:
            tmp[i] = vecs[i] * thetaTilde[i] + \
                     Jdx[i]  * ( yPD[i] * 1.0/( c[i] + s[i] ) + \
                                 yPDrange2[i] * 1.0/( cR[i] + s[i] ) )
        J.matvec_transp( tmp, Hvx )    # Hvx = J^T (Theta J dx + ThetaTilde ds)
        Hvx += self.hprod( -shifted_y, vecx )         # Lagrangian contribution
        for i in self.Bounds:
            Hvx[i] += xi[i] * vecx[i]                # Hvx += E^T Xi E dx

        # Components of t appear in increasing order of x-variables indices
        j = 0
        for i in range( self.n ):
            if i in self.lowerB + self.upperB + self.rangeB:
                Hvx[i] += xi[i] * vect[j]
                j += 1                               # Hvx += E^T Xi dt

        Hvs = numpy.zeros( self.nConst, 'd' )
        #for i in self.Eq + self.Ineq + self.Range:
        #    Hvs[i] = thetaTilde[i] * Jdx[i] + ( uPD[i]/s[i] + thetaTilde[i] ) * vecs[i]
        for i in self.Eq + self.lowerC:
            Hvs[i] = thetaTilde[i] * Jdx[i] + ( uPD[i]/s[i] + thetaTilde[i] ) * vecs[i]
        for i in self.upperC:
            Hvs[i] = thetaTilde[i] * Jdx[i] + ( uPD[i]/s[i] - thetaTilde[i] ) * vecs[i]
        for i in self.Range:
            Hvs[i] = thetaTilde[i] * Jdx[i] + \
                     ( uPD[i]/s[i] + (yPD[i] * 1.0/( c[i] + s[i] ) + \
                                      yPDrange2[i] * 1.0/( cR[i] + s[i] ) ) ) * vecs[i]

        Hvt = numpy.zeros( self.nBounds, 'd' )
        #print ' vecx = ', vecx
        #print ' vect = ', vect
        j = 0
        for i in range( self.n ):
            if i in self.lowerB + self.upperB + self.rangeB:
                Hvt[j] = xi[i] * vecx[i] + ( vPD[i]/t[i] + xi[i] ) * vect[j]
                j += 1
        

        return numpy.concatenate( (Hvx, Hvs, Hvt) )

    # ========================================================================

    def UnpackStep( self, step ):
        # Unpacks 'step' as returned e.g by PyGLTR into its x, s, t components
        newt = {}
        lenx = self.n
        lens = lenx + self.nConst

        # Components of t appear in increasing order of x-variables indices
        j = 0
        for i in range( self.n ):
            if i in self.lowerB + self.upperB + self.rangeB:
                newt[i] = step[lens+j]
                j += 1

        return ( step[:lenx], step[lenx:lens], newt )

    # ========================================================================

    def UpdateMu( self ):
        self.mu *= self.mufactor  # Simplistic for now

        # Update KKT tolerances
        self.dual_feas_stop = 1.1 * min( self.mu**1.0001, self.mu )
        self.CSY_compl_stop = 1.1 * self.mu
        self.SU_compl_stop  = 1.1 * self.mu
        self.XTZ_compl_stop = 1.1 * self.mu
        self.TV_compl_stop  = 1.1 * self.mu
        #if self.mu < 1.0e-5:
        #    self.dual_feas_stop = 1.0e-5
        #    self.CSY_compl_stop = 1.0e-5
        #    self.SU_compl_stop = 1.0e-5
        #    self.XTZ_compl_stop = 1.0e-5
        #    self.TV_compl_stop = 1.0e-5
        return

    # ========================================================================

    def PrimalFeasibility( self, x ):
        (c, cR) = self.evalC( x )
        # Compute || cE || and || cI- || with infinity norm
        normCE = normCI = 0.0
        for i in self.Eq:
            normCE = max(normCE, abs( c[i] ))
        for i in self.Ineq:
            normCI = max( normCI, max(-c[i], 0.0) )
        for i in self.Range:
            normCI = max( normCI, max(-cR[i], 0.0) )

        # Compute max violation of bounds
        violationX = 0.0
        for i in self.lowerB:
            violationX = max( violationX, self.Lvar[i] - x[i] )
        for i in self.upperB:
            violationX = max( violationX, x[i] - self.Uvar[i] )
        for i in self.rangeB:
            violationX = max(violationX, self.Lvar[i]-x[i], x[i]-self.Uvar[i])

        return( normCE, normCI, violationX )

    # ========================================================================

    def UpdateNu( self, x ):
        ( normCE, normCI, violationX ) = self.PrimalFeasibility( x )
        if normCE > self.etaE or normCI > self.etaI or violationX > self.etaX:
            self.nuEq = max( self.tau1 * self.nuEq, self.nuEq + self.tau2 )
            self.nuS  = max( self.tau1 * self.nuS,  self.nuS  + self.tau2 )
            self.nuT  = max( self.tau1 * self.nuT , self.nuT  + self.tau2 )

        self.etaE = min( max( 10 * self.mu**0.4, 1.0e-6 ), 0.2 )
        self.etaI = min( max( 10 * self.mu**0.4, 1.0e-6 ), 0.2 )
        self.etaX = min( max( 10 * self.mu**0.4, 1.0e-6 ), 0.2 )
        return

    # ========================================================================

    def ComputePDMult( self, x, s, t, step, y, yrange2, z, zrange2, u, v ):

        # Unpack step
        (stepx, steps, stept) = self.UnpackStep( step )

        # Obtain primal multipliers
        (yP, yrange2P, zP, zrange2P, uP, vP) = self.PrimalMult( x, s, t )

        # Compute J(x) dx
        (c, cR) = self.evalC( x )
        J = self.jac( x )
        Jdx = numpy.zeros( self.nConst, 'd' )
        J.matvec( stepx, Jdx )

        stepy = numpy.zeros( self.nConst, 'd' )
        stepyrange2 = {}
        stepu = numpy.zeros( self.nConst, 'd' )
        stepv = {}
        stepz = {}
        stepzrange2 = {}

        # Update multipliers y
        for i in self.Eq + self.lowerC + self.Range:
            stepy[i] = yP[i] -  y[i] - \
                       y[i]/(c[i] + s[i]) * ( steps[i] + Jdx[i] )
        for i in self.upperC:
            stepy[i] = yP[i] - y[i] - \
                       y[i]/(c[i] + s[i]) * ( steps[i] - Jdx[i] )
        for i in self.Range:
            stepyrange2[i] = yrange2P[i] - yrange2[i] - \
                             yrange2[i]/(cR[i] + s[i]) * (steps[i] - Jdx[i])

        # Update multipliers z
        for i in self.lowerB + self.rangeB:
            stepz[i] = zP[i] - z[i] - \
                       z[i]/(x[i] - self.Lvar[i] + t[i]) * (stept[i] + stepx[i])
        for i in self.upperB:
            stepz[i] = zP[i] - z[i] - \
                       z[i]/(self.Uvar[i] - x[i] + t[i]) * (stept[i] - stepx[i])
        for i in self.rangeB:
            stepzrange2[i] = zrange2P[i] - zrange2[i] - \
                             zrange2[i]/(self.Uvar[i] - x[i] + t[i]) * \
                             (stept[i] - stepx[i])

        # Update multipliers u
        for i in self.Eq + self.Ineq:
            stepu[i] = uP[i] - u[i] - u[i]/s[i] * steps[i]

        # Update multipliers v
        for i in self.Bounds:
            stepv[i] = vP[i] - v[i] - v[i]/t[i] * stept[i]

        return( stepy, stepyrange2, stepz, stepzrange2, stepu, stepv )

    # ========================================================================
    # Obtain fraction-to-the-boundary step length

    def get_alpha(self, x, s, t, step, y, yrange2, z, zrange2, u, v, stepy, stepyrange2, stepz, stepzrange2, stepu, stepv  ):

        alpha = 1.0

        (stepx, steps, stept) = self.UnpackStep( step )
        
        for i in self.Eq + self.Ineq:
            if stepy[i] < 0.0:
                alpha = min( alpha, -y[i]/stepy[i] )
            if stepu[i] < 0.0:
                alpha = min( alpha, -u[i]/stepu[i] )
            if steps[i] < 0.0:
                alpha = min( alpha, -s[i]/steps[i] )
        for i in self.Range:
            if stepy[i] < 0.0:
                alpha = min( alpha, -y[i]/stepy[i] )
            if stepyrange2[i] < 0.0:
                alpha = min( alpha, -yrange2[i]/stepyrange2[i] )
            if steps[i] < 0.0:
                alpha = min( alpha, -s[i]/steps[i] )	    	
        
        for i in self.Bounds:
            if stepz[i] < 0.0:
                alpha = min( alpha, -z[i]/stepz[i] )
            if stepv[i] < 0.0:
                alpha = min( alpha, -v[i]/stepv[i] )
            if stept[i] < 0.0:
                alpha = min( alpha, -t[i]/stept[i] )
        for i in self.rangeB:
            if stepz[i] < 0.0:
                alpha = min( alpha, -z[i]/stepz[i] )
            if stepzrange2[i] < 0.0:
                alpha = min( alpha, -zrange2[i]/stepzrange2[i] )
            if stept[i] < 0.0:
                alpha = min( alpha, -t[i]/stept[i] )
		
		
        feasible = self.CheckFeasibleLite( x, s, t, step, alpha )
        while not feasible:
            alpha /= 2
            feasible = self.CheckFeasibleLite( x, s, t, step, alpha )

        return alpha * self.ftb

    # ========================================================================

    def UpdatePDMult( self, x, s, t, step, y, yrange2, z, zrange2, u, v, alpha = 1.0 ):

        # Obtain updated primal-dual multipliers
        (stepy, stepyrange2, stepz, stepzrange2, stepu, stepv) = \
         self.ComputePDMult( x, s, t, step, y, yrange2, z, zrange2, u, v )

        # Update multipliers
        y_new = y + alpha * stepy
        yrange2_new = self.AddSparseVectors( yrange2, stepyrange2, alpha=alpha )
        z_new = self.AddSparseVectors( z, stepz, alpha = alpha )
        zrange2_new = self.AddSparseVectors( zrange2, stepzrange2, alpha=alpha )
        u_new = u + alpha * stepu
        v_new = self.AddSparseVectors( v, stepv, alpha=alpha )

        # Project primal-dual multipliers componentwise if requested
        #
        # ----> Should this all happen at (x,s,t) + alpha * step ??? <----
        #
        if self.project_mult:

            kappaL = self.kappaL
            kappaU = self.kappaU

            # Obtain primal multipliers at (x,s,t) + step
            (stepx, steps, stept) = self.UnpackStep( step )
            new_t = self.AddSparseVectors( t, stept )
            (yP, yrange2P, zP, zrange2P, uP, vP) = self.PrimalMult( x+stepx, s+steps, new_t )

            # Treat inequality constraints
            for i in self.Eq + self.Ineq + self.Range:
                L = kappaL * min( 1.0, y[i], yP[i] )
                U = max( kappaU, y[i], kappaU/self.mu, kappaU * yP[i] )
                y_new[i] = max( L, min( y_new[i], U ) )

                L = kappaL * min( 1.0, u[i], uP[i] )
                U = max( kappaU, u[i], kappaU/self.mu, kappaU * uP[i] )
                u_new[i] = max( L, min( u_new[i], U ) )            

            # Treat second part of range constraints
            for i in self.Range:
                L = kappaL * min( 1.0, yrange2[i], yrange2P[i] )
                U = max(kappaU, yrange2[i], kappaU/self.mu, kappaU*yrange2P[i])
                yrange2_new[i] = max( L, min( yrange2_new[i], U ) )

            # Treat bound constraints
            for i in self.lowerB + self.upperB + self.rangeB:
                L = kappaL * min( 1.0, z[i], zP[i] )
                U = max( kappaU, z[i], kappaU/self.mu, kappaU * zP[i] )
                z_new[i] = max( L, min( z_new[i], U ) )

                L = kappaL * min( 1.0, v[i], vP[i] )
                U = max( kappaU, v[i], kappaU/self.mu, kappaU * vP[i] )
                v_new[i] = max( L, min( v_new[i], U ) )

            # Treat second part of range bounds
            for i in self.rangeB:
                L = kappaL * min( 1.0, zrange2[i], zrange2P[i] )
                U = max(kappaU, zrange2[i], kappaU/self.mu, kappaU*zrange2P[i])
                zrange2_new[i] = max( L, min( zrange2_new[i], U ) )

        return ( y_new, yrange2_new, z_new, zrange2_new, u_new, v_new )

    # ========================================================================

    def InnerResiduals( self, x, s, t, y, yrange2, z, zrange2, u, v ):

        error_return = ( -1.0, -1.0, -1.0, -1.0, -1.0 )
        
        gPhiPD = self.GradPhiPD( x, s, t, y, yrange2, z, zrange2, u, v )
        residualG = self.InfinityNorm( gPhiPD )

        maxXT = 0.0
        residualCSY = 0.0
        residualSU  = 0.0
        residualXTZ = 0.0
        residualTV  = 0.0
        (c, cR) = self.evalC( x )
        (cps, cpsR) = self.CplusS( c, cR, s )
        
        for i in self.Ineq + self.Eq + self.Range:

            # Second block is (C+S)y - mu e
            if y[i] < 0.0: return error_return
            complCSYi = cps[i] * y[i] - self.mu
            residualCSY = max( residualCSY, abs( complCSYi ) )

            # Third block is Su - mu e
            if u[i] < 0.0: return error_return
            complSUi = s[i] * u[i] - self.mu
            residualSU = max( residualSU, abs( complSUi ) )

        for i in self.Range:
            if yrange2[i] < 0.0: return error_return
            complCSYi = cpsR[i] * yrange2[i] - self.mu
            residualCSY = max( residualCSY, abs( complCSYi ) )

        for i in self.lowerB:
            if z[i] < 0.0: return error_return
            if v[i] < 0.0: return error_return
            xti = x[i] - self.Lvar[i] + t[i]
            complXZi = xti * z[i] - self.mu
            residualXTZ = max( residualXTZ, abs( complXZi ) )
            maxXT = max( maxXT, xti )
            complTVi = t[i] * v[i] - self.mu
            residualTV = max( residualTV, abs( complTVi ) )

        for i in self.upperB:
            if z[i] < 0.0: return error_return
            if v[i] < 0.0: return error_return
            xti = self.Uvar[i] - x[i] + t[i]
            complXZi = xti * z[i] - self.mu
            residualXTZ = max( residualXTZ, abs( complXZi ) )
            maxXT = max( maxXT, xti )
            complTVi = t[i] * v[i] - self.mu
            residualTV = max( residualTV, abs( complTVi ) )
            
        for i in self.rangeB:
            xti= max( x[i] - self.Lvar[i] + t[i], self.Uvar[i] - x[i] + t[i] )
            if z[i] < 0.0: return error_return
            if zrange2[i] < 0.0: return error_return
            complXZi = max( abs( (x[i] - self.Lvar[i] + t[i]) * z[i] - self.mu ),
                            abs( (self.Uvar[i] - x[i] + t[i]) * zrange2[i] - self.mu ) )
            maxXT = max( maxXT, xti )
            residualXTZ = max( residualXTZ, complXZi )

        # Compute residuals scalings
        scalingG = max( 1.0, self.Max( y ), self.MaxD( yrange2 ) )

        scalingCSY = max( 1.0, self.Max( y ), self.Max( cps ), \
                          self.MaxD( cpsR ), self.MaxD( yrange2 ) )

        scalingSU = max( 1.0, self.Max( u ), self.Max( s ) )

        scalingXTZ = 1.0
        for i in z.keys():
            scalingXTZ = max( scalingXTZ, z[i], maxXT )
        for i in zrange2.keys():
            scalingXTZ = max( scalingXTZ, zrange2[i], maxXT )

        scalingTV = max( 1.0, self.MaxD( t ), self.MaxD( v ) )
        #print '      scaling factors: ', ( scalingG, scalingCSY, scalingSU, scalingXTZ, scalingTV )
        #print '      unscaled resids: ', ( residualG, residualCSY, residualSU, residualXTZ, residualTV )

        # Scale residuals
        residualG   /= scalingG
        residualCSY /= scalingCSY
        residualSU  /= scalingSU
        residualXTZ /= scalingXTZ
        residualTV  /= scalingTV

        return ( residualG, residualCSY, residualSU, residualXTZ, residualTV )
    
    # ========================================================================

    # Verifies whether residuals meet tolerance requirements

    def SufficientlySmall( self, res ):
        
        if (res[0] <= self.dual_feas_stop and
            res[1] <= self.CSY_compl_stop and
            res[2] <= self.SU_compl_stop  and
            res[3] <= self.XTZ_compl_stop and
            res[4] <= self.TV_compl_stop ):
            return True
        return False

    # ========================================================================

    # Check that a given point is feasible with respect to both
    # general constraints and simple bounds.
    # Return 1 if step is feasible, 0 if infeasible.

    def CheckFeasible( self, c, cR, x, s, t ):

        # Check feasibility with respect to general constraints
        for i in self.Eq + self.Ineq + self.Range:
            if ( c[i] + s[i] < 0.0 or s[i] < 0.0 ):
                return False

        for i in self.Range:
            if ( cR[i] + s[i] <= 0.0 or s[i] <= 0.0 ):
                return False

        # Check feasibility with respect to simple bounds
        for i in self.lowerB:
            if ( x[i] - self.Lvar[i] + t[i] <= 0.0 or t[i] <= 0.0 ):
                return False

        for i in self.upperB:
            if ( self.Uvar[i] - x[i] + t[i] <= 0.0 or t[i] <= 0.0 ):
                return False

        for i in self.rangeB:
            if ( x[i] - self.Lvar[i] + t[i] <= 0.0 or
                 self.Uvar[i] - x[i] + t[i] <= 0.0 or t[i] <= 0.0 ):
                return False

        # All is well, step is strictly feasible
        return True

    # =======================================================================    
    def CheckFeasibleLite( self, x, s, t, step, alpha ):

        (stepx, steps, stept) = self.UnpackStep( step )
        
        t_test = self.AddSparseVectors( t, stept, alpha = alpha )
        x_test = x + alpha*stepx
        s_test = s + alpha*steps
        (c, cR) = self.evalC( x_test )
           
        # Check feasibility with respect to general constraints
        for i in self.Eq + self.Ineq + self.Range:
            if (c[i] + s_test[i] <= 0.0 or s_test[i] <= 0.0):
                return False

        for i in self.Range:
            if (cR[i] + s_test[i] <= 0.0 or s_test[i] <= 0.0):
                return False

        # Check feasibility with respect to simple bounds
        for i in self.lowerB:
            if (x_test[i]-self.Lvar[i]+t_test[i] <= 0.0 or t_test[i] <= 0.0):
                return False
					
        for i in self.upperB:
            if (self.Uvar[i]-x_test[i]+t_test[i] <= 0.0 or t_test[i] <= 0.0):
                return False

        for i in self.rangeB:
            if (x_test[i] - self.Lvar[i] + t_test[i] <= 0.0 or
                 self.Uvar[i]-x_test[i]+t_test[i] <= 0.0 or t_test[i] <= 0.0):
                return False

        # All is well, step is strictly feasible
        return True

    # ========================================================================

    # Add two dictionaries, such as t, v or z
    # The two dictionaries are required to have the same keys for now
    # I will build a more complete Class later...
    
    def AddSparseVectors( self, v1, v2, alpha = 1.0 ):
        v = {}
        for k in v1.keys():
            v[k] = v1[k] + alpha * v2[k]
        return v

    # ========================================================================

    def InfinityNorm( self, v ):
        if len( v ) > 0:
            return numpy.linalg.norm( v, ord = numpy.Inf ) #max( abs( v ) )
        return self.negInfinity

    # ========================================================================

    # We only compute the max or min of positive vectors
    # Set the max or min to zero when the vector is void

    def Max( self, seq ):
        if len( seq ):
            return numpy.max( seq )
        else:
            return 0.0

    def Min( self, seq ):
        if len( seq ):
            return numpy.min( seq )
        else:
            return 0.0

    # Min or Max of a dictionnary containing nonnegative real numbers

    def MaxD( self, d ):
        m = self.negInfinity
        for k in d.keys():
            m = max( m, d[k] )
        return m

    def MinD( self, d ):
        m = self.Infinity
        for k in d.keys():
            m = min( m, d[k] )
        return m

    # ========================================================================

    # Return approximate value of machine epsilon

    def _Epsilon( self ):
        one = 1.0
        eps = 1.0
        while (one + eps) > one:
            eps = eps / 2.0
        return eps*2.0

########################################################################

########################################################################

# Test Class InteriorFramework

if __name__ == "__main__":
    import sys
    import pygltr
    import trustregion
    
    if len( sys.argv ) < 2:
        print 'Use %s <problem-name>' % sys.argv[0]
        sys.exit()
    # End if...
    ProblemName = sys.argv[1]

    # Main instantiation
    nlp = SuperbInteriorFramework( ProblemName, nu = 1.0, mu = 5.0 )

    x = nlp.x0
    s = nlp.s
    t = nlp.t
    n = nlp.n
    nConst = nlp.nConst

    print 'Details of problem %-s' % ProblemName
    print 'has %-d variables' % n
    print '        %-9d lower bounds   \t%-9d upper bounds  \t%-9d range bounds' \
          % (nlp.nlowerB, nlp.nupperB, nlp.nrangeB )
    print '        %-9d fixed variables\t%-9d free variables' \
          % (nlp.nfixedB, nlp.nfreeB )
    print '    %-d constraints' % nlp.m
    print '        %-9d equality constraints' % nlp.nEq
    print '        %-9d inequality constraints' % nlp.nIneq
    print '            %-9d with lower bound' % nlp.nlowerC
    print '            %-9d with upper bound' % nlp.nupperC
    print '            %-9d range constraints' % nlp.nrangeC
    print

    # Instantiate trust-region framework
    TR = trustregion.TrustRegionFramework( Delta = 1.0 )

    print ' ------------------'
    f = nlp.obj( x )
    (c, cR) = nlp.evalC( x )
    mins = nlp.Min( s )
    (cps, cpsR) = nlp.CplusS( c, cR, s )
    mincps = min( nlp.Min( cps ), nlp.MinD( cpsR ) )
    Phi = nlp.merit0
    gPhi = nlp.GradPhi( x, s, t )

    # Set inital primal-dual multipliers
    y = numpy.ones( nConst, 'd' )
    z = numpy.zeros( n, 'd' )
    v = numpy.zeros( n, 'd' )
    for i in nlp.lowerB + nlp.upperB + nlp.rangeB:
        z[i] = 1.0
        v[i] = 1.0
    yrange2 = numpy.ones( nlp.nRange, 'd' )
    zrange2 = numpy.ones( nlp.nrangeB, 'd' )
    u = numpy.ones( nConst, 'd' )

    print ' =================='
    n_outer = 0
    res = nlp.InnerResiduals( x, s, t, y, yrange2, z, zrange2, u, v )

    # Print header for outer iteration
    print ' %-6s\t%-10s\t%-10s\t%-10s\t%-10s\t%-10s\t%-10s\t%-8s\t%-8s' % \
          ('Outer', 'Obj', 'DualFeas', 'Comp-CSY', 'Comp-SU', 'Comp-XTZ', 'Comp-TV', 'Mu', 'Nu')
    print ' %-6d\t%-10g\t%-10g\t%-10g\t%-10g\t%-10g\t%-10g\t%-8g\t%-8g' % \
          (n_outer, f, res[0], res[1], res[2], res[3], res[4], nlp.mu, nlp.nu)

    HPhi = nlp.HessPhi( x, s, t, y, yrange2, z, zrange2, u, v )
    print ' =================='
    
    n_inner = 0
    rho = 1.0

    # Print header for inner iteration
    print '    %-5s\t%-15s\t%-10s\t%-10s\t%-6s\t%-10s\t%-2s\t%-10s\t%-10s\t%-10s' % \
          ( 'Inner', 'Merit', 'Delta', 'Step', '#Cg', 'Rho', 'Nc', 'TR-mult', 'min(s)', 'min(c+s)')

    # Begin inner iteration
    while not nlp.SufficientlySmall( res ):
        n_inner += 1

        (m, l, step_norm, step, cgiter, nc, ierr ) = pygltr.gltr( gPhi, HPhi, TR.Delta )

        if ierr:
            print '     Error exit from PyGLTR:: ierr  = ', ierr
        # Enf if...
        
        (stepx, steps, stept) = nlp.UnpackStep( step )
        
        c_trial = nlp.evalC( x + stepx )

        # Check for infeasible step
        infeasible = not nlp.CheckFeasible( c_trial, x + stepx, s + steps, t + stept )

        radius = TR.Delta
        if infeasible:
            print '     Warning:: Infeasible step, cutting back'
            TR.UpdateRadius( -1.0, step_norm )
        # End if infeasible

        if not infeasible:
            Phi_trial = nlp.Phi( x + stepx, s + steps, t + stept )

            rho = TR.Rho( Phi, Phi_trial, m )
            if rho == None:
                print '     Error:: Negative predicted reduction !'
            else:
                TR.UpdateRadius( rho, step_norm )
            # End if rho....

            if rho >= TR.eta1:
                x += stepx
                s += steps
                t += stept
                (y, yrange2, z, zrange2, u, v) = nlp.UpdatePDMult( x, s, t, y, yrange2, z, zrange2, u, v )
                HPhi = nlp.HessPhi( x, s, t, y, yrange2, z, zrange2, u, v )
                Phi  = Phi_trial
                gPhi = nlp.GradPhi( x, s, t )
                f    = nlp.obj( x )
                print '     ++ Step accepted ++'
                mins   = nlp.Min( s )
                (cps, cpsR) = nlp.CplusS( c, cR, s )
                mincps = min( nlp.Min( cps ), nlp.MinD( cpsR ) )

                res = nlp.InnerResiduals( x, s, t, y, yrange2, z, zrange2, u, v )
                print '   Residuals: ', res

            # End If rho...
        # End if not infeasible

        # Print some info on current inner iteration
        print '    %-5d\t%-15.6g\t%-10.6g\t%-10.6g\t%-6d\t%-10.6g\t%-2d\t%-10g\t%-10g\t%-10g' % \
              ( n_inner, Phi, radius, step_norm, cgiter, rho, nc, l, mins, mincps )

    # End While not...
    # End inner iteration

    print ' ------------------'
    print ' All residuals small'
    print ' Outer iteration finished in %-d inner iterations' % n_inner
    print ' ------------------'
    nlp.UpdateMu()
    nlp.UpdateNu( x )
    print
    print ' %-6s\t%-10s\t%-10s\t%-10s\t%-10s\t%-10s\t%-10s\t%-8s\t%-8s' % \
          ('Outer', 'Obj', 'DualFeas', 'Comp-CSY', 'Comp-SU', 'Comp-XTZ', 'Comp-TV', 'Mu', 'Nu')
    print ' %-6d\t%-10g\t%-10g\t%-10g\t%-10g\t%-10g\t%-10g\t%-8g\t%-8g' % \
          (n_outer, f, res[0], res[1], res[2], res[3], res[4], nlp.mu, nlp.nu)

    print ' Final x: ', x

    nlp.close()
    sys.exit()

# End if __name__ == ...
