"""
Sparse tools for AmplPy and PySparse.

 . coord2ll() converts a sparse matrix in coordinate format
   to sparse linked-list format.

 , find() performs the inverse operation.

 . spygnu() plots the sparsity pattern of a sparse matrix in
   coordinate format using Gnuplot.

 . spymat() is similar, using matplotlib (a must!).

 . spygnull() plots the sparsity pattern of a sparse matrix in
   sparse linked-list format using Gnuplot.

 . spymatll() is simlilar, using matplotlib.

$Id$
"""

##############################################################################

####                                                                      ####
#               M. Friedlander and D. Orban, for AmplPy                      #
#               Chicago, May 2003.                                           #
####                                                                      ####

##############################################################################

import _sparsetools
from pysparse import spmatrix
import numpy

##############################################################################

def FastSpy(nrow, ncol, irow, jcol, **kwargs):
    import pylab
    ax = kwargs.get( 'ax', None )
    if ax is None:
        fig = pylab.figure()
        ax = fig.gca()
    sym = kwargs.get( 'sym', False )
    t = kwargs.get('title', None)
    comment = kwargs.get('comment', '')
    showtitle = kwargs.get('showtitle', True)

    nnz = len(irow)
    ms = 1
    if nrow+ncol <= 100: ms = 5
    ax.plot(jcol, irow, 'ks', markersize=ms, linestyle='None')
    if sym:
        ax.plot(irow, jcol, 'ks', markersize=ms, linestyle='None')
    ax.xaxis.set_major_locator(pylab.MaxNLocator(nbins=9,
                                                 steps=[1, 2, 5, 10],
                                                 integer=True))
    ax.yaxis.set_major_locator(pylab.MaxNLocator(nbins=9,
                                                 steps=[1, 2, 5, 10],
                                                 integer=True))
    ax.xaxis.set_ticks_position('both')
    ax.yaxis.set_ticks_position('both')
    ax.set_xlim(xmin=-1, xmax=ncol)
    ax.set_ylim(ymin=nrow, ymax=-1)
    ax.set_aspect('equal')
    if nrow == ncol:
        #ax.set_aspect('equal')
        if t is None: t = 'Order %-d with %-d nonzeros' % (nrow,nnz)
    else:
        #ratio = (1.0*nrow)/ncol
        #ax.set_aspect(ratio)
        if t is None: t = 'Size (%-d,%-d) with %-d nonzeros' % (nrow,ncol,nnz)
    if t is not None: t = comment + t
    if showtitle: ax.set_title( t, fontsize='small' )
    return ax

def spymat( nrow, ncol, irow, icol, **kwargs ):
    """
    Plots the sparsity pattern of sparse matrix using the matplotlib
    plotting package. The matrix should be in sparse coordinate
    format, as returned by find().
    
    Recognized keyword arguments are
    
        comment    a comment to be appended to the title of the plot
                    (default: no comment),
        sym        should be set to True if the matrix is symmetric and
                   only its lower or upper triangle is stored,
        val        if not None, will be used as color code for the magnitude
                   of nonzero elements. A good idea is to pass the values of
                   the elements of the matrix.
        ax         a set of axes in which to plot the sparsity pattern.
                   If not specified, a new figure will be created.

    Returns a handle to the plot and the newly created axes if applicable.
    See the matplotlib documentation for more information.
    """
    
    val = kwargs.get( 'val', None )
    sym = kwargs.get( 'sym', False )
    comment = kwargs.get( 'comment', '' )
    ax = kwargs.get( 'ax', None )

    ret2 = False
    if ax is None:
        import matplotlib
        if matplotlib.__version__ < '0.65':
            import matplotlib.matlab as MM
        else:
            import matplotlib.pylab as MM
        ax = MM.axes()
        ret2 = True

    # Monitor nnz
    nnz1 = len( irow )
    # If matrix is not square, it cannot be symmetric
    if nrow != ncol :
        sym = False
    if sym:
        nnz = 2 * nnz1
    else:
        nnz = nnz1

    rowind = numpy.zeros( nnz )
    colind = numpy.zeros( nnz )
    if val is not None:
        mag = 100.0/max(abs( numpy.asarray(val) ))
        s = numpy.zeros( nnz, 'd' )
        intensities = s
        minmaxstr = ' (min,max) = (%-g,%-g)' % ( min( val ), max( val ) )
    else:
        intensities = 'k'
        s = 1
        minmaxstr = ''

    # Account for (0,0) in NW corner in axes definition
    for i in range( len( irow ) ):
        colind[i] = icol[i]
        rowind[i] = irow[i]
        if val is not None: s[i] = abs(val[i]) * mag
    # If matrix is symmetric, include transpose of lower triangle
    for i in range( nnz1, nnz ):
        colind[i] = irow[i-nnz1]
        rowind[i] = icol[i-nnz1]
        if val is not None: s[i] = abs(val[i-nnz1]) * mag

    handle = ax.scatter( colind, rowind, s, intensities, marker = 'o', alpha = 0.75 )
    ax.set_xlim( [-1,ncol] )
    ax.set_ylim( [nrow,-1] )

    if nrow == ncol:
        ax.set_aspect('equal')
        t = 'Order %-d with %-d nonzeros' % (nrow,nnz)
    else:
        ax.set_aspect()
        t = 'Size (%-d,%-d) with %-d nonzeros' % (nrow,ncol,nnz)
    t = comment + t + minmaxstr
    ax.set_title( t, fontsize='small' )

    if ret2:
        return (ax, handle)
    else:
        return handle
    
##############################################################################

def spygnu( nrow, ncol, irow, icol, sym=0, term='x11', comment='', outfile=None ):
    """
    Plots the sparsity pattern of sparse matrix 'matrix' using the gnuplot
    plotting package. The 'matrix' argument should be in sparse coordinate
    format, as returned by find(). The optional argument 'comment'
    will be appended to the title of the plot, 'term' should be set to a
    valid gnuplot terminal value and 'outfile' to a file name, if appropriate
    for the value of 'term'. 'sym' should be set to a nonzero value if the
    matrix is symmetric and only its lower or upper triangle is stored.
    """
    import tempfile, os
    import math
    import Gnuplot, Gnuplot.funcutils

    # Monitor nnz
    nnz = len( irow )
    if sym:
        nnz = 2 * nnz
        # If matrix is not square, it cannot be symmetric
        if nrow != ncol :
            sym = 0
    
    # Open temp file
    tmpname  = tempfile.mktemp()
    template = open( tmpname, 'w' )
    
    # Write out sparsity pattern -- account for origin in NW corner
    for i in range( len( irow ) ):
        template.write( "%d\t%d\n" % ( icol[i], nrow-irow[i] ) )
        if sym:
            template.write( "%d\t%d\n" % ( irow[i], ncol-icol[i] ) )
            if icol[i] == irow[i]:
                nnz = nnz-1    # Not to count the diagonal twice
    template.close()
    # Set up Gnuplot
    g = Gnuplot.Gnuplot( persist=1, debug=0 )
    g.title( comment + ' (%d X %d) Sparsity pattern, nnz = %d' % (nrow, ncol, nnz ) )
    g( 'set nokey' )
    if term == 'ps':
        g( 'set terminal postscript eps color solid "Helvetica" 12' )
    if outfile is not None:
        g( 'set output "' + outfile + '"' )
    else:
        if term == 'ps':
            output_tmp = tempfile.mktemp()
            g( 'set output "' + output_tmp + '"' )
    g( 'set xrange[0:%d]' % ncol )
    g( 'set yrange[0:%d]' % nrow )
    # Must reverse the labels on the y axis
    xstep = max( pow( 10, int( math.log10( max( ncol, 1 ) ) ) )/2, 1 )
    ystep = max( pow( 10, int( math.log10( max( nrow, 1 ) ) ) )/2, 1 )
    g( 'set tics out' )
    g( 'set xtics 0,%d,%d' % ( xstep,  ncol ) )
    g( 'set ytics %d,%d,0' % ( nrow, -ystep ) )
    g( 'set noautoscale' )
    if ncol == nrow:
        g( 'set size square' )
    else:
        g( 'set size ratio %d/%d' % (nrow, ncol) )
    # Set point size according to matrix size
    pointsize = min( 1.0/max( math.log10( max( nrow, ncol ) + 1 ), 1.0 ), 1.0 )
    apply( g.plot, [ '"' + tmpname + '"' + ' with points ps ' + str( pointsize ) ] )
    g.gnuplot.gnuplot.close()
    os.remove( tmpname )

#############################################################################

def spygnull( matrix, comment='', term = 'x11', outfile=None ):
    """
    Plots the sparsity pattern of sparse matrix 'matrix' using the gnuplot
    plotting package. The 'matrix' argument should be in sparse linked-list
    format, as available in package spmatrix. The optional argument 'comment'
    will be appended to the title of the plot, 'term' should be set to a
    valid gnuplot terminal value and 'outfile' to a file name, if appropriate
    for the value of 'term'.
    """
    (val, irow, icol) = find( matrix )
    (nrow, ncol) = matrix.shape
    spygnu( nrow, ncol, irow, icol,
         sym=matrix.issym, comment=comment, term=term, outfile=outfile )

#############################################################################

def spymatll( matrix, color = None, **kwargs ):
    """
    Plots the sparsity pattern of sparse matrix 'matrix' using the matplotlib
    plotting package. The 'matrix' argument should be in sparse linked-list
    format, as available in package spmatrix. The optional argument 'color'
    defaults to None. If not None, a color will be assigned to elements on the
    plot based on their magnitude. The optional argument 'comment'
    will be appended to the title of the plot.
    Returns a handle to the plot. See the matplotlib documentation.
    """
    (values, irow, icol) = find( matrix )
    (nrow, ncol) = matrix.shape
    if color is not None:
        val = values
    else:
        val = None
    #return spymat( nrow, ncol, irow, icol, val=val, sym=matrix.issym, **kwargs )
    return FastSpy( nrow, ncol, irow, icol, val=val, sym=matrix.issym, **kwargs )

#############################################################################

def coord2ll( matrix, nrow, ncol, nnz, sym=0 ):
    """
    Convert a sparse matrix in coordinate format to linked-list format.
    The argument matrix should be a triple (data, irow, icol). nnz is
    the number of nonzero elements in the matrix, or in its upper
    triangle if sym is set to a nonzero value.
    """
    data = matrix[0]
    irow = matrix[1]
    icol = matrix[2]
    if sym:
        M = spmatrix.ll_mat_sym( nrow, nnz )
    else:
        M = spmatrix.ll_mat( nrow, ncol, nnz )
        
    for i in range( nnz ):
        M[ irow[i], icol[i] ] = data[i]
   
    return M

#############################################################################

def find( M ):
    """
    Return a triple (val, irow, icol) of numarrays pointing to the nonzero
    elements of the input matrix M. The nonzero element val[i] of the sparse
    matrix M is located at the intersection of row irow[i] and column icol[i].
    M can be given in either LL or CSR format.
    """
    if type( M ).__name__ == 'll_mat':
        return _sparsetools.find_ll( M )
    if type( M ).__name__ == 'csr_mat':
        return _sparsetools.find_csr( M )
    else:
        return None

#############################################################################

def rdm( n, m = -1, nnz = -1, sym = False, NoEmpty = True ):
    """
    Generate a random sparse matrix in LL format of the specified dimensions.
      rdm( n )
            generates a random nonsymmetric n x n matrix.
    Keywords:
      m        specify the number of columns (default: m = n)
      nnz      specify the number of nonzero entries (default: nnz = 2n)
      sym      specify whether matrix should be symmetric (default: sym = False)
      NoEmpty  specify that matrix must have no empty row or column (default: NoEmpty = True)
    """
    if m == -1: m = n
    if nnz == -1: nnz = 2*n
    if NoEmpty and nnz < max(n, m):
        raise ValueError, ' rdm():: must have nnz >= max(n,m) if NoEmpty = True'
        return None
    import types
    if type(n) != types.IntType or type(m) != types.IntType or type(nnz) != types.IntType:
        raise TypeError, ' rdm():: n, m and nnz must be integer'
        return None
    if n != m and sym:
        raise ValueError, ' rdm():: matrix must be square if symmetric'
        return None
    import random
    if sym:
        nnz /= 2
        new = spmatrix.ll_mat_sym( n, nnz )
    else:
        new = spmatrix.ll_mat( n, m, nnz )
    ilist = []
    jlist = []
    for elem in range(nnz):
        goti = False
        while not goti:
            i = random.randint( 0, n-1 )
            if not i in ilist: goti = True
        gotj = False
        while not gotj:
            if sym: j = random.randint( 0, i )
            else: j = random.randint( 0, m-1 )
            if not j in jlist: gotj = True
        mag = random.randint( 1, 100 )
        new[(i,j)] = mag * random.random()
    return new 

#############################################################################

def usolve( U, rhs ):
    """
    x = usolve( U, rhs )
    Solve a nonsingular upper triangular system of equations U x = rhs.
    If U is not in CSR format, it will be converted prior to solving the
    system of equations.
    """
    n, m = U.shape
    if n != m:
        raise TypeError, 'U must be square'
        return None
    if rhs.shape != (n,):
        raise TypeError, 'rhs does not have appropriate length'
        return None
    x = numpy.zeros( n, 'd' )
    if type(U).__name__ == 'll_mat':
        _sparsetools.usolve_csr( U.to_csr(), rhs, x )
    elif type(U).__name__ == 'csr_mat':
        _sparsetools.usolve_csr( U, rhs, x )
    else:
        raise TypeError, 'Unrecognized matrix type'
        return None
    return x

#############################################################################

def utsolve( U, rhs ):
    """
    x = usolve( Ut, rhs )
    Solve a nonsingular lower triangular system of equations U' x = rhs,
    where U is upper triangular. If U is not in CSR format, it will be
    converted prior to solving the system of equations.
    """
    n, m = U.shape
    if n != m:
        raise TypeError, 'U must be square'
        return None
    if rhs.shape != (n,):
        raise TypeError, 'rhs does not have appropriate length'
        return None
    x = numpy.zeros( n, 'd' )
    if type(U).__name__ == 'll_mat':
        _sparsetools.utsolve_csr( U.to_csr(), rhs, x )
    elif type(U).__name__ == 'csr_mat':
        _sparsetools.utsolve_csr( U, rhs, x )
    else:
        raise TypeError, 'Unrecognized matrix type'
        return None
    return x

#############################################################################

def lsolve( L, rhs ):
    """
    x = lsolve( L, rhs )
    Solve a nonsingular lower triangular system of equations L x = rhs.
    If L is not in CSR format, it will be converted prior to solving the
    system of equations.
    """
    n, m = L.shape
    if n != m:
        raise TypeError, 'L must be square'
        return None
    if rhs.shape != (n,):
        raise TypeError, 'rhs does not have appropriate length'
        return None
    x = numpy.zeros( n, 'd' )
    if type(L).__name__ == 'll_mat':
        _sparsetools.lsolve_csr( L.to_csr(), rhs, x )
    elif type(L).__name__ == 'csr_mat':
        _sparsetools.lsolve_csr( L, rhs, x )
    else:
        raise TypeError, 'Unrecognized matrix type'
        return None
    return x

#############################################################################

def ltsolve( L, rhs ):
    """
    x = ltsolve( L, rhs )
    Solve a nonsingular upper triangular system of equations L' x = rhs,
    where L is a lower triangular matrix. If L is not in CSR format, it
    will be converted prior to solving the system of equations.
    """
    n, m = L.shape
    if n != m:
        raise TypeError, 'L must be square'
        return None
    if rhs.shape != (n,):
        raise TypeError, 'rhs does not have appropriate length'
        return None
    x = numpy.zeros( n, 'd' )
    if type(L).__name__ == 'll_mat':
        _sparsetools.ltsolve_csr( L.to_csr(), rhs, x )
    elif type(L).__name__ == 'csr_mat':
        _sparsetools.ltsolve_csr( L, rhs, x )
    else:
        raise TypeError, 'Unrecognized matrix type'
        return None
    return x

#############################################################################

def uppertriangularize( M ):
    n = M.shape[0]
    M.generalize()

    # Make sure there are no zeros on diagonal
    for i in range(n):
        if M[i,i] == 0.0:
            M[i,i] = 1.0

    # Remove lower triangular part of matrix
    for i in range(n):
        for j in range(i):
            M[i,j] = 0.0

#############################################################################

def lowertriangularize( M ):
    n = M.shape[0]
    M.generalize()

    # Make sure there are no zeros on diagonal
    for i in range(n):
        if M[i,i] == 0.0:
            M[i,i] = 1.0

    # Remove upper triangular part of matrix
    for i in range(n):
        for j in range(i+1,n):
            M[i,j] = 0.0

#############################################################################

def Hilbert( n ):
    # The cream of ill conditioning: the Hilbert matrix.  See Higham,
    # "Accuracy and Stability of numpyal Algoriths", section 28.1.
    # The matrix has elements H(i,j) = 1/(i+j-1) when indexed
    # i,j=1..n.  However, here we index as i,j=0..n-1, so the elements
    # are H(i,j) = 1/(i+j+1).
    if n <= 0: return None
    if n == 1: return 1.0
    nnz = n * (n - 1)/2
    H = spmatrix.ll_mat_sym( n, nnz )
    for i in range( n ):
        for j in range( i+1 ):
            H[i,j] = 1.0/(i+j+1)
    return H
