#!/usr/bin/env python
#
#      - - - - - >[ T R I P ]- - - - - >
#
"""
A Trust-Region Interior-Point method for nonlinear
programming. This version only accepts problems
with inequality constraints and bounds.

This module demonstrates features of NLPy. The
barrier method in itself is purely primal, and
no attempt was made to make it as fast or as
sophisticated as possible.
D. Orban                       Montreal Sept. 2003

$Id:$
"""

import amplpy         # Used to load in models
import numpy
import getopt, sys    # Used for I/O and to parse cmd line
import generic_io     # Generic input/output
import pygltr         # Used to solve trust-region subproblems
import trustregion
import norms
import trip_class as TC       # Auxilliary functions

explicit = True       # Use explicit or implicit solves in PyGltr
verbose  = False
ny = True             # Perform Nocedal-Yuan backtracking on trust-region reject

## TRIP starts here
ProblemName = generic_io.parse_cmdline(sys.argv[1:])

# Instantiate an interior-point framework
nlp = TC.InteriorFramework( ProblemName )

# Check that problem has no equalities or range constraints
if nlp.nEq > 0 or nlp.nrangeC > 0:
    print ' Problem %-s has %-d equality constraints' % (ProblemName, nlp.nEq)
    print ' and %-d range inequalities' % nlp.nrangeC
    print ' TRIP only accepts problems without equalities or range constraints'
    nlp.close()
    sys.exit( 1 )

# Initialize counters
iter = 0
cgiter = 0

# Initial data
x = nlp.x0

# Move x into the bounds
for i in range( nlp.n ):
    x[i] = max( nlp.Lvar[i] + 0.01, min( nlp.Uvar[i] - 0.01, x[i] ) )

f = nlp.obj( x )
f0 = f
phi = nlp.Phi( x )
g = nlp.GradPhi( x )
gnorm = norms.norm2( g )    # Initial gradient norm
gnorm0 = gnorm
infeasibility = nlp.Infeasibility( x )
TR = trustregion.TrustRegionFramework()
rho = 1               # Dummy initial value for rho
status = ''           # Dummy initial step status

if verbose:
    print ' lowerC = ', nlp.lowerC
    print ' upperC = ', nlp.upperC
    print ' lowerB = ', nlp.lowerB
    print ' upperB = ', nlp.upperB
    print ' rangeB = ', nlp.rangeB
    print ' x = ', x
    print ' f = ', f
    print ' J = ', nlp.evalJ( x )
    print ' g = ', g
    print ' H = ', nlp.HessPhi( x )

print
print '------------------------------------------'
print 'Trip. Solving problem %-s with parameters' % ProblemName
print 'eta1 = %-g  eta2 = %-g  gamma1 = %-g  gamma2 = %-g' % ( TR.eta1, TR.eta2, TR.gamma1, TR.gamma2 )
print '------------------------------------------'
print

banner_fmt = "%-5s  %11s  %11s  %10s  %8s  %6s  %-4s"
data_fmt   = "%-5d  %11.4e  %11.4e  %10.4e  %8.2e  %6.2f  %-4s"
banner = banner_fmt % ( 'Iter', 'f(x)', 'Phi(x,mu)', '||KKTd||', 'mu', 'rho', 'Stat' )
lbanner = len(banner)
print '-' * lbanner
print banner
print '-' * lbanner
print data_fmt % ( iter, f, phi, gnorm, nlp.mu, rho, status )

while nlp.mu > 1.0e-7 and gnorm > max( 1.0e-8, gnorm0 * 1.0e-5 ):
    TR.Delta = max( 1.0, 0.1 * gnorm )
    while gnorm > 10 * nlp.mu:
        iter += 1
        # Iteratively minimize the quadratic model in the trust region
        # m(s) = <g, s> + 1/2 <s, Hs>
        # Note that m(s) does not include f(x): m(0) = 0.
        feasible = False
        while not feasible:
            G = pygltr.PyGltrContext( g, radius = TR.Delta )
            if explicit:
                G.explicit_solve( nlp.HessPhi( x ) )
            else:
                G.implicit_solve( lambda v: nlp.HessPhiProd( x, v ) )
            m = G.m
            z = G.mult
            ierr = G.ierr
            snorm = G.snorm
            if ierr != 0:
                sys.stderr.write('Pygltr terminates with exit code %-d\n' % ierr)
            infeasibility = nlp.Infeasibility( x + G.step )
            if infeasibility > 0.0:
                TR.UpdateRadius( -1.0, snorm )
                if verbose:
                    print ' Infeasible; adjusting radius to ', TR.Delta
            else:
                feasible = True
        cgiter += G.niter
        x_trial = x + G.step
        phi_trial = nlp.Phi( x_trial )
        rho = TR.Rho( phi, phi_trial, m )
        TR.UpdateRadius( rho, snorm )
        status = 'Rej'
        if rho >= TR.eta1:
            x = x_trial
            phi = phi_trial
            f = nlp.obj( x )
            g = nlp.GradPhi( x )
            gnorm = norms.norm2( g )
            infeasibility = nlp.Infeasibility( x )
            status = 'Acc'
        else:
            if ny: # Backtracking linesearch following "Nocedal & Yuan"
                alpha = 1.0
                slope = numpy.dot( g, G.step )
                while phi_trial >= phi + 1.0e-4 * alpha * slope:
                    alpha /= 2
                    x_trial = x + alpha * G.step
                    phi_trial = nlp.Phi( x_trial )
                x = x_trial
                phi = phi_trial
                f = nlp.obj( x )
                g = nlp.GradPhi( x )
                gnorm = norms.norm2( g )
                TR.Delta = alpha * G.snorm
                status = 'N-Y'
        print data_fmt % ( iter, f, phi, gnorm, nlp.mu, rho, status )

    # End inner while

    # Compute gradient with updated value of mu and loop
    nlp.UpdateMu()
    phi = nlp.Phi( x )
    g = nlp.GradPhi( x )
    gnorm = norms.norm2( g )
    infeasibility = nlp.Infeasibility( x )

# End outer while

print '-' * 80
if verbose:
    print ' Final x = ', x
    print ' Final mu = ', nlp.mu
    (zc,zx) = nlp.PrimalMult( x )
    print ' zc = ', zc
    print ' zx = ', zx

# Output some final stats
print
print '-----------------------'
print 'Trip: End of execution'
print '  Problem..................... : %-s' % ProblemName
print '  Number of variables......... : %-d' % nlp.n
print '  Number of constraints....... : %-d' % (nlp.nlowerC + nlp.nupperC)
print '  Number of bounds............ : %-d' % nlp.nBounds
print '  Initial/Final objective..... : %-g/%-g' % ( f0, f )
print '  Initial/Final gradient norm. : %-g/%-g' % ( gnorm0, gnorm )
print '  Final infeasibility......... : %-g' % infeasibility
print '  Number of iterations........ : %-d' % iter
print '  Total/Average Lanczos iter.. : %-g/%-g' % ( cgiter, float(cgiter)/iter )
print '-----------------------'

# Output "solution"
#nlp.writesol( x0, pi0, 'And the winner is' )

# Close connection with model
nlp.close()
