## Various functions
#$Id:$

import math

def usage( n ):
    sys.stderr.write( "Use\n %s problem" % Caller )
    sys.stderr.write( "where problem is unconstrained..." )
    exit( n )

def commandline_err( msg ):
    sys.stderr.write( "%s: %s\n" % ( Caller, msg ) )
    sys.exit( 1 )

def parse_cmdline( arglist ):
    if len( arglist ) != 1:
        commandline_err( 'Specify file name' )
        return None

    try: options, fname = getopt.getopt( arglist, '' )

    except getopt.error, e:
        commandline_err( "%s" % str( e ) )
        return None

    return fname[0]

def norm2( x ):
    n2x = 0.0
    for i in range( len( x ) ):
        n2x += pow( x[i], 2.0 )
    return math.sqrt( n2x )

########################

class InteriorFramework:

    """
    ipm = InteriorFramework( problem )
    Instantiate an InteriorFramework class embedding the nonlinear
    program whose name is 'problem'. As part of the embedding, the
    bound constraints are reformulated as follows
    
        xL <= x <= xU      becomes       x  - xL >= 0, and
                                         xU - x  >= 0
        xL <= x            becomes       x  - xL >= 0
        x  <= xU           becomes       xU - x  >= 0.

    The counters nrangeB, nlowerB, nupperB countain the number of bounds
    of the first, second and third kind respectively. To those are
    appended the number of fixed and free variables: nfixedB and nfreeB.
    After reformulations, there are
        nBounds = 2 * nrangeB + nlowerB + nupperB
    'effective' bound constraints.

    Similarly, constraints are reformulated as follows

        cL <= c(x) <= cU   .......       are currently *ignored*,
        cL <= c(x)         becomes       c(x) - cL   >= 0
              c(x) <= cU   becomes       cU   - c(x) >= 0.

    The counters nrangeC, nlowerC and nupperC contain the number of
    constraints of the first, second and third kind respectively. To those
    are appended the number of equality and 'free' constraints (if any):
    nequalC and nfreeC. The effective number of inequality constraints
    after reformulation s thus
        nIneq = nlowerC + nupperC.
    """

    import amplpy
    import math

    # Constructor
    def __init__( self, ProblemName ):
        self.nlp = amplpy.AmplModel( ProblemName )

        # Ignore range constraints for now
        Ineq    = lowerC + upperC
        Eq      = equalC
        
        # Maintain counters for effective number of constraints and bounds
        nBounds = 2 * nrangeB + nlowerB + nupperB
        nIneq   = nlowerC + nupperC
        nEq     = nequalC

    # Destructor
    def close( self ):
        self.nlp.close()

    ### Class methods ###

    # Reformulate constraints.
    # For now, exclude range constraints.
    def evalC( self, x ):
        c = self.nlp.cons( x )
        j = 0
        for i in self.nlp.rangeC:
            pass
        for i in self.nlp.lowerC:
            c[i] -= self.nlp.Lcon[i]
        for i in self.nlp.upperC:
            c[i] = self.nlp.Ucon[i] - c[i]
        return c

    def evalJ( self, x ):
        J = self.nlp.jac( x )
        # Reverse sign of gradient for 'upper' inequalities
        for i in range( self.nlp.nupperC ):
            J[i,:] = J[i,:].times_scalar( -1 )        
        return J

    # Proceed similarly with bound constraints
    # Want to transform bounds into x >= 0;
    # Order them using: [ range, lower, upper, fixed, free ].
    def evalBnds( self, x ):
        b = Numeric.zeros( self.nBounds , 'd' )
        j = 0
        for i in range( 0, self.nlp.nrangeB, 2 ):
            b[i]   = self.nlp.x[i] - self.nlp.Lvar[i]
            b[i+1] = self.nlp.Uvar[i] - self.nlp.x[i]
        j += 2 * self.nlp.nrangeB
        for i in range( self.nlp.nlowerB ):
            b[j + i] = self.nlp.x[i] - self.nlp.Lvar[i]
        j += self.nlp.nlowerB
        for i in range( self.nlp.nupperB ):
            b[j + i] = self.nlp.Uvar[i] - self.nlp.x[i]
        j += self.nlp.nupperB
        for i in range( self.nlp.nfixedB + self.nlp.nfreeB ):
            b[j + i] = self.nlp.x[i]
        return b
        
    # Evaluate the barrier function and its derivatives
    # phi(x, mu) = f(x) - mu * Sum log( ci(x) ) - mu * Sum log( xi )
    def Phi( self, x, mu ):
        f = self.nlp.obj( x )
        c = self.evalC( nlp, x )
        b = self.evalBnds( x )
        phi = f
        for i in self.Ineq:
            phi -= mu * math.log( c[i] )
        for i in range( self.nBounds ):
            phi -= mu * math.log( b[i] )
        return phi

    # grad phi(x, mu) = grad f(x) - mu * Sum 1/ci(x) grad ci(x) - mu * Sum 1/xi ei
    def GradPhi( self, x, mu ):
        g = self.nlp.grad( x )
        c = self.evalC( x )
        J = self.evalJ( x )
        gPhi = g
        for i in self.Ineq:
            gPhi -= mu * 1.0/c[i] * J[:,i]
        for i in range( self.nlp.nrangeB ):
            gPhi[i] -= mu * ( 1.0/b[i] + 1.0/b[i+1] )
        for i in range( self.nlp.nlowerB ):
            gPhi[nIneq+2*nrangeB+i] -= mu * 1.0/b[2*nrangeB+i]
        for i in range( self.nlp.nupperB ):
            gPhi[nIneq+2*nrangeB+nlowerB+i] -= mu * 1.0/b[2*nrangeB+nlowerB+i]
        return gPhi

    # Evaluate matrix-vector products with the primal-dual Hessian of the barrier
    # H = H(x,z) + J(x)^T C^{-1}(x) Z J(x)
    def HessPhiProd( self, x, z, mu, v ):
        # Evaluate constraints and Jacobian
        c = self.evalC( x )
        J = self.evalJ( x )
        # Include contribution of Jacobian
        #    J(x)^T C^{-1}(x) Z J(x) v
        Hv = Numeric.zeros( nlp.n, 'd' )
        J.matvec( v, Hv )
        for i in range( nlp.m ):
            Hv[i] = Hv[i] * z[i] / c[i]
        J.matvec_transp( Hv, Hv )
        # Add product  H(x,z) * v
        Hv += self.nlp.hprod( z, v )

