#############################
#                           #
# Class definition for TRIP #
#                           #
#############################

import numpy
import math
from amplpy import AmplModel
from pysparse import spmatrix

# Class InteriorFramework inherits from AmplModel

class InteriorFramework( AmplModel ):

    """
    ipm = InteriorFramework( problem )
    Instantiate an InteriorFramework class embedding the nonlinear
    program whose name is 'problem'. As part of the embedding, the
    bound constraints are reformulated as follows
    
        xL <= x <= xU      becomes       x  - xL >= 0, and
                                         xU - x  >= 0
        xL <= x            becomes       x  - xL >= 0
        x  <= xU           becomes       xU - x  >= 0.

    The counters nrangeB, nlowerB, nupperB countain the number of bounds
    of the first, second and third kind respectively. To those are
    appended the number of fixed and free variables: nfixedB and nfreeB.
    After reformulations, there are
        nBounds = 2 * nrangeB + nlowerB + nupperB
    'effective' bound constraints.

    Similarly, constraints are reformulated as follows

        cL <= c(x) <= cU   .......       are currently *ignored*,
        cL <= c(x)         becomes       c(x) - cL   >= 0
              c(x) <= cU   becomes       cU   - c(x) >= 0.

    The counters nrangeC, nlowerC and nupperC contain the number of
    constraints of the first, second and third kind respectively. To those
    are appended the number of equality and 'free' constraints (if any):
    nequalC and nfreeC. The effective number of inequality constraints
    after reformulation s thus
        nIneq = nlowerC + nupperC.

    $Id:$
    """

    # Constructor
    def __init__( self, ProblemName, mu = 1.0 ):
        AmplModel.__init__( self, ProblemName )

        self.Bounds  = self.lowerB + self.upperB + self.rangeB

        # Ignore range constraints for now
        self.Ineq    = self.lowerC + self.upperC
        self.Eq      = self.equalC
        
        # Maintain counters for effective number of constraints and bounds
        self.nBounds = 2 * self.nrangeB + self.nlowerB + self.nupperB
        self.nIneq   = self.nlowerC + self.nupperC
        self.nEq     = self.nequalC

        self.mu      = mu
        self.mu_factor = 0.5   # mu+ = mu * mu_factor
        
        # Terminate outer iteration when norm of gradient of barrier is less than
        self.stop_absolute = 1.0e-6
        
        # Terminate inner iteration when norm of gradient of barrier is less than
        # self.stop_tol * self.mu
        self.stop_tol = 1.0

    # Destructor: inherited from AmplModel

    ### Class methods ###

    # Reformulate constraints.
    # For now, ignore range constraints.
    def evalC( self, x ):
        c = self.cons( x )
        for i in self.lowerC:
            c[i] -= self.Lcon[i]
        for i in self.upperC:
            c[i] = self.Ucon[i] - c[i]
        return c

    def Infeasibility( self, x ):
        c = self.cons( x )
        infeas = 0.0
        for i in self.lowerC:
            infeas = max( infeas, self.Lcon[i] - c[i] )
        for i in self.upperC:
            infeas = max( infeas, c[i] - self.Ucon[i] )
        for i in self.lowerB:
            infeas = max( infeas, self.Lvar[i] - x[i] )
        for i in self.upperB:
            infeas = max( infeas, x[i] - self.Uvar[i] )
        for i in self.rangeB:
            infeas = max( max( infeas, x[i] - self.Uvar[i] ), self.Lvar[i] - x[i] )
        return infeas

    def evalJ( self, x ):
        return self.jac( x )

    # Evaluate the barrier function and its derivatives
    # phi(x, mu) = f(x) - mu * Sum log( ci(x) ) - mu * Sum log( xi )
    def Phi( self, x ):
        f = self.obj( x )
        c = self.evalC( x )
        phi = f
        for i in self.Ineq:
            phi -= self.mu * math.log( c[i] )
        for i in self.lowerB + self.rangeB:
            phi -= self.mu * math.log( x[i] - self.Lvar[i] )
        for i in self.upperB + self.rangeB:
            phi -= self.mu * math.log( self.Uvar[i] - x[i] )
        return phi

    # Compute the primal multipliers
    def PrimalMult( self, x ):
        c = self.evalC( x )
        
        # Primal multipliers corresponding to general constraints
        zc = numpy.zeros( self.nlowerC + self.nupperC, 'd' )
        for i in self.lowerC:
            zc[i] = self.mu/c[i]
        for i in self.upperC:
            zc[i] = -self.mu/c[i]

        # Primal multipliers corresponding to bound constraints
        zx = {}
        for i in self.lowerB:
            zx[i] = self.mu/(x[i] - self.Lvar[i])
        for i in self.upperB:
            zx[i] = -self.mu/(self.Uvar[i] - x[i])
        for i in self.rangeB:
            zx[i] = self.mu * ( 1.0/(x[i] - self.Lvar[i]) - 1.0/(self.Uvar[i] - x[i]) )

        return (zc, zx)

    # grad phi(x, mu) = grad f(x) - mu * Sum 1/ci(x) grad ci(x) - mu * Sum 1/xi ei
    def GradPhi( self, x ):
        (zc, zx) = self.PrimalMult( x )
        gPhi = numpy.zeros( self.n, 'd' )
        J = self.evalJ( x )
        J.matvec_transp( -zc, gPhi )
        gPhi += self.grad( x )
        for i in self.lowerB + self.upperB + self.rangeB:
            gPhi[i] -= zx[i]
        return gPhi

    def HessPhi( self, x ):
        (zc,zx) = self.PrimalMult( x )
        c = self.evalC( x )
        J = self.evalJ( x )
        J2 = spmatrix.ll_mat( J.shape[0], J.shape[1], J.nnz )
        for i in self.lowerC:
            J2[i,:] = J[i,:].times_scalar( zc[i]/c[i] )
        for i in self.upperC:
            J2[i,:] = J[i,:].times_scalar( -zc[i]/c[i] )
        H = spmatrix.dot( J, J2 )
        H.shift( 1.0, self.hess( x, -zc ) )
        for i in self.lowerB:
            H[i,i] += zx[i]/(x[i] - self.Lvar[i])
        for i in self.upperB:
            H[i,i] -= zx[i]/(self.Uvar[i] - x[i])
        for i in self.rangeB:
            H[i,i] += self.mu * ( 1.0/(x[i] - self.Lvar[i])**2 + 1.0/(self.Uvar[i] - x[i])**2 )
        return H

    def HessPhiProd( self, x, v ):
        (zc,zx) = self.PrimalMult( x )
        c = self.evalC( x )
        J = self.evalJ( x )
        Hv = numpy.zeros( self.n, 'd' )
        tmp = numpy.zeros( self.nlowerC + self.nupperC, 'd' )
        J.matvec( v, tmp )
        for i in self.lowerC:
            tmp[i] *= zc[i]/c[i]
        for i in self.upperC:
            tmp[i] *= -zc[i]/c[i]
        J.matvec_transp( tmp, Hv )
        Hv += self.hprod( -zc, v )
        for i in self.lowerB:
            Hv[i] += zx[i]/(x[i] - self.Lvar[i]) * v[i]
        for i in self.upperB:
            Hv[i] -= zx[i]/(self.Uvar[i] - x[i]) * v[i]
        for i in self.rangeB:
            Hv[i] += self.mu * ( 1.0/(x[i] - self.Lvar[i])**2 + 1.0/(self.Uvar[i] - x[i])**2 ) * v[i]
        return Hv

    def UpdateMu( self ):
        self.mu *= self.mu_factor
        #self.mu = (self.mu)**2/(2*self.mu + 10)
        return

# Test Class InteriorFramework

if __name__ == "__main__":
    import sys
    ProblemName = sys.argv[1]
    nlp = InteriorFramework( ProblemName )
    print 'Details of problem %-s' % ProblemName
    print '    %-d variables' % nlp.n
    print '        %-d lower bounds' % nlp.nlowerB
    print '        %-d upper bounds' % nlp.nupperB
    print '        %-d range bounds' % nlp.nrangeB
    print '        %-d fixed variables' % nlp.nfixedB
    print '        %-d free  variables' % nlp.nfreeB
    print '    %-d constraints' % nlp.m
    print '        %-d equality   constraints' % nlp.nEq
    print '        %-d inequality constraints' % nlp.nIneq
    print '            %-d with lower bound' % nlp.nlowerC
    print '            %-d with upper bound' % nlp.nupperC
    print '            %-d range constraints' % nlp.nrangeC

    lastn = min( nlp.n, 4 )
    lastm = min( nlp.m, 4 )
    
    (x0, z0) = (nlp.x0, nlp.pi0)
    b = nlp.evalBnds( x0 )

    # Move x0 inside the bounds
    for i in range( nlp.n ):
        if i in nlp.lowerB:
            x0[i] = b[i] + 0.1
        elif i in nlp.upperB:
            x0[i] = b[i] - 0.1
        elif i in nlp.rangeB:
            x0[i] = 0.5 * ( b[i] + b[i+1] )

    b = nlp.evalBnds( x0 )

    z0 = numpy.ones( nlp.m, 'd' )
    print ' x0 = ', x0
    print ' z0 = ', z0
    print ' b  = ', b
    print ' f( x0 ) = ', nlp.obj( x0 )
    c = nlp.evalC( x0 )
    print ' c( x0 ) = ', c

    feasible = 1
    for i in nlp.Ineq:
        if c[i] <= 0.0:
            print ' x0 is not strictly feasible!'
            feasible = 0
            break

    if feasible:
        for i in range( nlp.nBounds ):
            if b[i] <= 0.0:
                print ' x0 does not satisfy the bounds strictly!'
                feasible = 0
                break

    if feasible:
        print ' mu = ', self.mu
        print ' phi( x0, mu ) = ', nlp.Phi( x0 )
        print ' GradPhi( x0, mu ) = ', nlp.GradPhi( x0 )
    
    nlp.close()

