#!/usr/bin/env python
"""
                    T R U N K

 A first unconstrained optimization solver in Python
 The Python version of the celebrated  F90/95 solver
 D. Orban                        Montreal Sept. 2003

 $Id:$
"""

import numpy        # Used to manipulate arrays
import amplpy         # Used to load in models
import sys
#import trpcg
#import ppcg
import math           # Used for various math operations
import generic_io
import trustregion    # Trust-Region management
from math import sqrt
from nlpy_timing import cputime

silent = False         # Print intermediate info
showbanner = True     # Print initial info

t = cputime()

## Trunk starts here

ProblemName = generic_io.parse_cmdline(sys.argv[1:]) 

# Create a model
nlp = amplpy.AmplModel( ProblemName )

# Check for unconstrained problem
if nlp.m != 0:
    sys.stderr.write( 'Problem has general constraints\n' )
    generic_io.usage( -1 )

stop_tol = 5.0 * 1e-6
cgtol = 1
iter = 0
cgiter = 0
niter = 0
x = nlp.x0
f = nlp.obj( x )
f0 = f
g = nlp.grad( x )

gnorm = sqrt(numpy.dot(g,g)) # Initial gradient norm
TR = trustregion.TrustRegionFramework( Delta = 0.1 * gnorm,
                                       eta1 = 0.05,
                                       eta2 = 0.9,
                                       gamma1 = 0.25,
                                       gamma2 = 2.5 )
rho = 1                  # Dummy initial value for rho
status = ''              # Dummy initial step status
solved = False

if showbanner:
    print
    print '------------------------------------------'
    print 'Trunk. Solving problem %-s with parameters' % ProblemName
    print 'eta1 = %-g  eta2 = %-g  gamma1 = %-g  gamma2 = %-g Delta0 = %-g' % \
           ( TR.eta1, TR.eta2, TR.gamma1, TR.gamma2, TR.Delta )
    print '------------------------------------------'
    print

    header = "%-4s  %12s  %12s  %8s  %8s  %5s  %8s  %4s" % ( 'Iter', 'f(x)', '||g(x)||', 'rho', 'Delta', 'CG', 'CGtol', 'Stat' )
    hline  = '-' * len(header)
    fmt = "%-4d  %12.5e  %12.5e  %8.1e  %8.1e  %5d  %8.1e  %3s"

t_setup = cputime() - t    # Setup time
t = cputime()

while not solved:

    # Print out header, say, every 20 iterations
    if not silent and iter % 20 == 0:
        print header
        print hline
    
    if not silent:
        print fmt % ( iter, f, gnorm, rho, TR.Delta, niter, cgtol, status )
        #print iter, f, gnorm, rho, TR.Delta, niter, cgtol, status

    if gnorm < stop_tol:
        solved = True
    else:
        # Iteratively minimize the quadratic model in the trust region
        # m(s) = <g, s> + 1/2 <s, Hs>
        # Note that m(s) does not include f(x): m(0) = 0.

        # Pass a matrix-vector product function to trpcg
        # Use inexact Newton stopping criterion
        # Will be multiplied by gnorm in trpcg
        cgtol = max( 1.0e-6, min( 0.5 * cgtol, sqrt(gnorm) ))
        #cgtol = 1.0e-6
        #(step, niter, snorm ) = trpcg.trpcg(
        #                            g,
        #                            Delta=TR.Delta,
        #                            hprod = lambda p: nlp.hprod(nlp.pi0,p),
        #                            tol = cgtol )
        #CG = ppcg.Ppcg( g,
        #                matvec = lambda p: nlp.hprod(nlp.pi0, p),
        #                radius = TR.Delta,
        #                reltol = cgtol,
        #                debug = False )
        #CG.Solve()
        #step = CG.x
        #niter = CG.iter
        #snorm = sqrt( CG.xNorm2 )

        TrSolver = trustregion.TrustRegionGLTR( g,
                                              matvec = lambda p: nlp.hprod(nlp.pi0,p),
                                              radius = TR.Delta,
                                              reltol = cgtol )
        TrSolver.Solve()
        step = TrSolver.step
        niter = TrSolver.niter
        snorm = TrSolver.stepNorm

        m = numpy.dot(g,step) + 0.5*numpy.dot(step, nlp.hprod(nlp.pi0,step))

        cgiter += niter
        x_trial = x + step
        f_trial = nlp.obj( x_trial )
        rho  = TR.Rho( f, f_trial, m )
        TR.UpdateRadius( rho, snorm )
        status = 'Rej'    # Step is rejected
        if rho >= TR.eta1:
            x = x_trial
            f = f_trial
            g = nlp.grad( x )
            gnorm = sqrt(numpy.dot(g,g))
            status = 'Acc'  # Step is accepted
        iter += 1

t_solve = cputime() - t    # Solve time

# Output some final statistics
print
print '-----------------------'
print 'Trunk-CG: End of execution'
print '  Problem              : %-s' % ProblemName
print '  Dimension            : %-d' % nlp.n
print '  Initial objective    : %-g' % f0
print '  Final   objective    : %-g' % f
print '  Final gradient norm  : %-g' % gnorm
print '  Number of iterations : %-d' % iter
print '  Total  CG iterations : %-g' % cgiter
print '  Average CG/iteration : %-g' % (float(cgiter)/iter)
print '  Setup time           : %-gs' % t_setup
print '  Solve time           : %-gs' % t_solve
print '  Total time           : %-gs' % (t_setup + t_solve)
print '-----------------------'

# Output "solution"
nlp.writesol( x, nlp.pi0, 'And the winner is' )

# Close connection with model
nlp.close()
